''' opencos.deps.deps_processor -- module is less about "parsing" and more about "processing"

a DEPS markup files targets (applying deps, reqs, commands, tags, incdirs, defines, etc) to a
CommandDesign ref object
'''

# pylint: disable=too-many-lines

import argparse
import copy
import glob
import os
from pathlib import Path

from opencos import files
from opencos import eda_config
from opencos.util import Colors, debug, info, warning, error, read_tokens_from_dot_f, \
    patch_args_for_dir, load_env_file
from opencos.utils import status_constants
from opencos.utils.str_helpers import dep_str2list
from opencos.utils.dict_helpers import dict_diff
from opencos.deps.deps_file import deps_target_get_deps_list
from opencos.deps.deps_commands import deps_commands_handler

from opencos.deps.defaults import SUPPORTED_TARGET_TABLE_KEYS, SUPPORTED_TAG_KEYS, \
    SUPPORTED_DEP_KEYS_BY_TYPE

class DepsProcessor: # pylint: disable=too-many-instance-attributes,too-many-public-methods
    '''DepsProcessor -- called by eda_base resolve_target_core(..)

    example usage:
    my_dp = DepsProcessor(
        command_design_ref = self, # aka, your CommandDesign obj
        deps_entry = <some-table-found-in-deps-data>,
        target = <target-str-we-were-looking-for>
        target_path = <filepath, for debug>
        target_node = <target leaf str, for debug>
        deps_file = <original DEPS.[markup-ext] for debug>
        caller_info = <str info for debug>
    )

    '''

    def __init__(self, command_design_ref, deps_entry: dict, target: str,
                 target_path: str, target_node: str, deps_file: str, caller_info: str):
        '''
        command_design_ref (eda.CommandDesign),
        deps_entry (dict, target in DEPS.yml file)
        target_node (str) -- key in DEPS.yml that got us the deps_entry, used for debug
        deps_file (str) -- file, used for debug
        caller_info (str) -- used for debug
        '''

        self.command_design_ref = command_design_ref
        self.deps_entry = deps_entry
        self.target = target
        self.target_path = target_path
        self.target_node = target_node # for debug
        self.deps_file = deps_file # for debug
        self.deps_dir, _ = os.path.split(deps_file)
        self.caller_info = caller_info

        # Check if it's a Command instead of CommandDesign:
        self.is_command_design = bool(
            getattr(command_design_ref, 'process_plusarg', None) and \
            getattr(command_design_ref, 'set_parameter', None) and \
            isinstance(getattr(command_design_ref, 'incdirs', None), list)
        )


        assert isinstance(deps_entry, dict), \
            f'{deps_entry=} for {target_node=} in {deps_file=} must be a dict'
        assert command_design_ref is not None, \
            'called DepsProcessor.__init__, but no ref to CommandDesign object (is None)'

        # named eda commands in the target:
        # If this deps_entry has a 'sim', 'build', etc command entry for this target, grab that
        # because it can set defines or other things specific to an eda command ('sim', for example)
        self.entry_eda_command = self.deps_entry.get(command_design_ref.command_name, {})

        # alias some of the self.command_design_ref values
        self.command_name = self.command_design_ref.command_name # str, for debug
        self.args         = self.command_design_ref.args         # dict
        self.config       = self.command_design_ref.config       # dict
        self.set_arg      = self.command_design_ref.set_arg      # method
        self.error        = self.command_design_ref.error        # method.

        # If there are expanded eda commands in
        # self.command_design_ref.config['command_handler'].keys(), then make note of that now.
        self.known_eda_commands = getattr(
            self.command_design_ref, 'config', {}
        ).get('command_handler', {}).keys()

        self._is_ok_deps_field_type(
            item=self.entry_eda_command, want_types=dict, name=self.command_name,
            msg='command entry in target'
        )


    def _defines_as_dict(self, defines) -> dict:
        '''Converts a dict/list/str defines obj to dict'''
        if isinstance(defines, dict):
            return defines

        if isinstance(defines, list):
            as_list = defines
        elif isinstance(defines, str):
            as_list = defines.split()
        else:
            assert False, \
                f'unknown type for at {self.caller_info=}: {defines=} {type(defines)=}'

        ret = {}
        for x in as_list:
            if x:
                if x.startswith('+define+'):
                    x = x[8:]
                parts = x.split('=')
                if parts[1:]:
                    ret[parts[0]] = '='.join(parts[1:])
                else:
                    ret[parts[0]] = None
        return ret


    def _plusargs_as_dict(self, plusargs) -> dict:
        '''Converts a dict/list/str plusargs obj to dict'''
        if isinstance(plusargs, dict):
            return plusargs

        if isinstance(plusargs, list):
            as_list = plusargs
        elif isinstance(plusargs, str):
            as_list = plusargs.split()
        else:
            assert False, \
                f'unknown type for at {self.caller_info=}: {plusargs=} {type(plusargs)=}'

        ret = {}
        for x in as_list:
            if x:
                if x.startswith('+'):
                    x = x[1:]
                parts = x.split('=')
                if parts[1:]:
                    ret[parts[0]] = '='.join(parts[1:])
                else:
                    ret[parts[0]] = ''
        return ret


    def _parameters_as_dict(self, parameters) -> dict:
        '''Converts a dict/list/str parameters obj to dict'''
        if isinstance(parameters, dict):
            return parameters

        if isinstance(parameters, list):
            as_list = parameters
        elif isinstance(parameters, str):
            as_list = parameters.split()
        else:
            assert False, \
                f'unknown type for at {self.caller_info=}: {parameters=} {type(parameters)=}'

        ret = {}
        for x in as_list:
            if x:
                if x.startswith('-G'):
                    x = x[2:]
                parts = x.split('=')
                if parts[1:]:
                    ret[parts[0]] = '='.join(parts[1:])
                else:
                    self.error(
                        f'At {self.caller_info}: parameter entry (non-Table, as markup',
                        f'Array/string) is not in the form Name=Value: "{x}"',
                       '\n    To see detailed DEPS schema, run:',
                       f'{Colors.byellow}eda deps-schema [--verbose]',
                       error_code=status_constants.EDA_DEPS_BAD_TYPE
                    )
        return ret


    def apply_defines(self, defines_dict: dict, check: bool = True) -> None:
        '''Given defines_dict, applies them to our self.command_design_ref obj'''
        if not self.is_command_design or not defines_dict:
            return
        if check and not self._is_ok_deps_field_type(defines_dict, (dict, list, str),
                                                     name='defines'):
            return
        if not isinstance(defines_dict, dict):
            defines_dict = self._defines_as_dict(defines_dict)

        for k,v in defines_dict.items():
            if v is None or v == '':
                self.command_design_ref.process_plusarg(f'+define+{k}')
            else:
                # %PWD% and %SEED% substiutions:
                if v and isinstance(v, str):
                    if v.startswith('%PWD%/') or v.startswith('"%PWD%/'):
                        v = v.replace('%PWD%', os.path.abspath(self.target_path))
                    if v.startswith('%SEED%') or v.startswith('"%SEED%'):
                        v = v.replace('%SEED%', str(self.args.get('seed', 1)))
                    # ENV var replacement:
                    v = os.path.expandvars(v)
                self.command_design_ref.process_plusarg(f'+define+{k}={v}')


    def apply_plusargs(self, plusargs_dict: dict, check: bool = True) -> None:
        '''Given plusarsg_dict, applies them to our self.command_design_ref obj'''
        if not self.is_command_design or not plusargs_dict:
            return
        if check and not self._is_ok_deps_field_type(plusargs_dict, (dict, list, str),
                                                     name='plusargs'):
            return

        if not isinstance(plusargs_dict, dict):
            plusargs_dict = self._plusargs_as_dict(plusargs_dict)

        for k,v in plusargs_dict.items():
            if v is None or v == '':
                self.command_design_ref.process_plusarg(f'+{k}')
            else:
                self.command_design_ref.process_plusarg(f'+{k}={v}')


    def apply_parameters(self, parameters_dict: dict, check: bool = True) -> None:
        '''Given parameters_dict, applies them to our self.command_design_ref obj'''
        if not self.is_command_design:
            return
        if check and not isinstance(parameters_dict, dict):
            self.error(f"{parameters_dict=} is not type dict, can't apply defines,",
                       f"in {self.caller_info}",
                       '\n    To see detailed DEPS schema, run:',
                       f'{Colors.byellow}eda deps-schema [--verbose]',
                       error_code=status_constants.EDA_DEPS_BAD_TYPE)
        for k,v in parameters_dict.items():
            if v is None or v == '' or not isinstance(v, (int, str, bool)):
                warning(f'parameter {k} has value: {v}, parameters must be bool/int/string types',
                        f'from {self.caller_info}')
                continue

            # since we are dealing with markup or a split markup string, we need to check
            # for integral types that may have been wrapped in strings:
            if isinstance(v, str):
                try:
                    v = int(v, 0)
                except ValueError:
                    pass
            self.command_design_ref.set_parameter(
                name=k, value=v, caller_info=self.caller_info
            )


    def apply_incdirs(self, incdirs_list:list) -> None:
        '''Given incdirs_list, applies them to our self.command_design_ref obj'''
        if not self.is_command_design:
            return
        if not isinstance(incdirs_list, (str, list)):
            self.error(f"{incdirs_list=} is not type str/list, can't apply incdirs",
                       f"in {self.caller_info}",
                       '\n    To see detailed DEPS schema, run:',
                       f'{Colors.byellow}eda deps-schema [--verbose]',
                       error_code=status_constants.EDA_DEPS_BAD_TYPE)

        incdirs_list = dep_str2list(incdirs_list)
        for x in incdirs_list:
            # Support both relative path to this DEPS file (current target) AND
            # if there is ENV var replacement. Try ENV var replacement first
            x_expanded = os.path.expandvars(x)
            if x_expanded and Path(x_expanded).is_absolute():
                # If abspath already, don't join it with DEPS file path:
                abspath = str(Path(x_expanded).resolve())
            elif x_expanded:
                # It's relative to DEPS file:
                abspath = os.path.abspath(os.path.join(self.target_path, x))
            else:
                # else, don't add blank incdirs
                abspath = ''

            if abspath and not os.path.exists(abspath):
                # Do warn if abspath doesn't exist
                warning(f'Unable to resolve incdir "{x}" ({abspath} does not exist)',
                        f'from {self.caller_info}')
            elif abspath and abspath not in self.command_design_ref.incdirs:
                self.command_design_ref.incdirs.append(abspath)
                debug(f'Added include dir "{abspath}" from {self.caller_info}')


    def _apply_args_check_tools(self, tokens: list, tagname: str) -> list:
        '''Helper for apply_args(list), returns list strips --tool args under certain conditions

        Basically, we want to see if a DEPS target want to set arg --tool, if so:
        Accept it:
          - this is not a tool class
          - tool was automatically chosen by eda.py's auto-tool-order (--tool not set at
             CLI)
          - accepting means set self.args['tool'], so we can respawn with this.
        Reject/Warn if:
          - previous tool was not automatically applied (--tool was set at CLI)
            and you're trying to change the --tool value.
          - This can happen if you have complicated DEPS targets trying to set the
            tool to different values.
        '''

        parser = argparse.ArgumentParser(
            prog='deps_processor --tool', add_help=False, allow_abbrev=False
        )
        parser.add_argument('--tool', default='')
        try:
            parsed, unparsed = parser.parse_known_args(tokens + [''])
            tokens2 = list(filter(None, unparsed))
        except argparse.ArgumentError:
            error('deps_processor --tool problem attempting to parse_known_args for:',
                  f'{tokens}, {self.caller_info} {tagname}')
            tokens2 = tokens

        _tool_class = 'tool' in self.args
        _orig_tool = self.args.get('tool', '')

        if not self.command_design_ref.auto_tool_applied and \
           _tool_class and parsed.tool and _orig_tool \
           and parsed.tool != _orig_tool:
            # tool arg present, --tool in this DEPS args, and tool already set.
            warning(
                f'Attempting to set --tool {parsed.tool} from DEPS',
                f'(file={self.deps_file}:{self.target_node})',
                f'however the tool was already chosen as: {_orig_tool}. The --tool arg will',
                f'not be applied from: {tokens}'
            )
        elif (self.command_design_ref.auto_tool_applied or not _tool_class) and parsed.tool:
            # tool arg wasn't present (not a Tool class), or it was auto-applied,
            # then add the arg anyway so we can later respawn with the correct tool.
            self.args['tool'] = parsed.tool
            debug(f'setting arg.tool to {parsed.tool=} from {self.caller_info}')

        # remove blanks, '--tool[=value| value]' removed.
        return [item for item in tokens2 if item != '']


    def apply_args( # pylint: disable=too-many-locals,too-many-branches,too-many-statements
            self, args_list:list, tagname: str = ''
    ) -> list:
        '''Given args_list, applies them to our self.command_design_ref obj

        This will return unparsed args that weren't in the self.command_design_ref.args keys
        unparsed args will show up as eda.py warnings, but will not fail. Most callers do not
        use the unparsed args from this method.
        '''
        if tagname:
            tagname = f'{tagname=}'

        if not isinstance(args_list, (str, list)):
            self.error(f"{args_list=} is not type str/list, can't apply args",
                       f"in {self.caller_info} {tagname}",
                       '\n    To see detailed DEPS schema, run:',
                       f'{Colors.byellow}eda deps-schema [--verbose]',
                       error_code=status_constants.EDA_DEPS_BAD_TYPE)

        prev_args = copy.deepcopy(self.args)

        tokens = dep_str2list(args_list)

        # patch args relative to the DEPS (if self.deps_dir exists) so things like
        # --build-tcl=<file> for relative <file> works when calling targets from any directory.
        tokens = patch_args_for_dir(
            tokens=tokens, patch_dir=self.deps_dir, caller_info=self.caller_info
        )

        # We're going to run an ArgumentParser here, which is not the most efficient
        # thing to do b/c it runs on all of self.command_design_ref.args (dict) even
        # if we're applying a single token.

        # Since some args (util.py, eda_config.py, eda.py) can only be handled from command
        # line, it would be nice if -f or --input-file is handled from DEPS, so we'll special
        # case that now. Recursively resolve -f / --input-file.
        # Do similary for --env-file (also only supported in util.py)
        parser = argparse.ArgumentParser(
            prog='deps_processor -f/--input-file', add_help=False, allow_abbrev=False
        )
        parser.add_argument('--env-file', default=[], action='append',
                            help=(
                                "dotenv file(s) to pass ENV vars, (default: .env loaded first,"
                                " subsequent files' vars override .env"
                            ))
        parser.add_argument('-f', '--input-file', default=[], action='append',
                            help=(
                                'Input .f file to be expanded as eda args, defines, incdirs,'
                                ' files, or targets.'
                            ))
        try:
            parsed, unparsed = parser.parse_known_args(tokens + [''])
            tokens2 = list(filter(None, unparsed))
        except argparse.ArgumentError:
            error('deps_processor -f/--input-file, problem attempting to parse_known_args for:',
                  f'{tokens}, {self.caller_info} {tagname}')
            tokens2 = tokens

        if parsed.input_file:
            dotf_tokens = []
            for filepath in parsed.input_file:
                # Since this isn't command line, we have to assume the path is relative
                # to this DEPS file.
                if not os.path.isabs(filepath):
                    filepath = os.path.join(self.deps_dir, filepath)
                dotf_tokens.extend(read_tokens_from_dot_f(
                    filepath=filepath, caller_info=self.caller_info, verbose=True
                ))

            # put the .f files before the unparsed args.
            tokens2 = dotf_tokens + tokens2

            # recurse until we've resolved nested .f files.
            return self.apply_args(args_list=tokens2)

        if parsed.env_file:
            for env_file in parsed.env_file:
                load_env_file(env_file)

        # if no --input-file/--env-file values, keep parsing the remaining tokens2:
        tokens = tokens2

        # We have to special-case anything with --tool[=value] in tokens, otherwise
        # the user may think they were allowed to set --tool, but in our flow the Command handler
        # (self.command_design_ref) has already been chosen, so setting the tool can have
        # strange side-effects.
        _tool_class = 'tool' in self.args
        _orig_tool = self.args.get('tool', '')
        tokens = self._apply_args_check_tools(tokens=tokens, tagname=tagname)

        debug(f'deps_processor - custom apply_args with {tokens=}',
              f'from {self.caller_info} {tagname}')
        _, unparsed = self.command_design_ref.run_argparser_on_list(
            tokens=tokens
        )

        # Annoying, but check for plusargs in unparsed, and have referenced CommandDesign
        # or CommandSim class handle it with process_plusarg.
        for arg in list(unparsed):
            if arg.startswith('+') and self.is_command_design:
                self.command_design_ref.process_plusarg(plusarg=arg, pwd=self.target_path)
                unparsed.remove(arg)

        # For any leftover files, or targets, attempt to process those too:
        for arg in list(unparsed):
            # Since this isn't command line, we have to assume for files, the path is relative
            # to this DEPS file.
            target = self.correct_a_deps_target(target=arg, deps_dir=self.deps_dir)

            file_exists, fpath, forced_extension = files.get_source_file(target)
            if file_exists:
                _, file_ext = os.path.splitext(fpath)
                if forced_extension or file_ext:
                    self.command_design_ref.add_file(fpath, caller_info=self.caller_info,
                                                     forced_extension=forced_extension)
                    unparsed.remove(arg)

            else:
                if not os.path.isdir(target) and \
                   self.is_command_design and \
                   self.command_design_ref.resolve_target_core(
                       target=target, no_recursion=False, caller_info=self.caller_info,
                       error_on_not_found=False
                   ):
                    unparsed.remove(arg)

        if unparsed:
            # This is only a warning - because things like CommandFlist may not have every
            # one of their self.args.keys() set for a given target, such as a 'sim' target that
            # has --optimize, which is not an arg for CommandFlist. But we'd still like to get an
            # flist from that target.
            warning(f'For {self.command_design_ref.command_name}:' \
                    + f' in {self.caller_info} has unknown args {unparsed=}')

        if (self.command_design_ref.auto_tool_applied or not _tool_class) and \
            _orig_tool != self.args.get('tool', ''):
            # If there was an auto tool applied (tool class or not) then attempt to pick
            # a new sub-command-object with that tool.
            debug(f'deps_processor.apply_args: tool changed, {self.args["tool"]=}, will attempt',
                  f'to respawn the job using original args: {self.config["eda_original_args"]}')
            self.command_design_ref.tool_changed_respawn = {
                'tool': self.args['tool'],
                'orig_tool': _orig_tool,
                'from': self.caller_info,
            }

        diff_args = dict_diff(prev_args, self.args)
        if diff_args:
            args_list = [item for item in args_list if item != ''] # remove blanks
            info(f'{Colors.off}{Colors.yellow}{self.caller_info} {tagname}{Colors.info}:',
                 f'applying args for {args_list}: {Colors.cyan}{diff_args}')

        return unparsed

    def apply_reqs(self, reqs_list:list) -> None:
        '''Given reqs_list, applies them ot our self.command_design_ref obj'''
        for req in reqs_list:
            req_abspath = os.path.abspath(os.path.join(self.target_path, req))
            self.command_design_ref.add_file(
                req_abspath, use_abspath=False, add_to_non_sources=True,
                caller_info=self.caller_info
            )

    def process_deps_entry( # pylint: disable=too-many-branches
            self
    ) -> list:
        '''Main entry point (after creating DepsProcessor obj) to resolve a deps target

        Example usage:
        deps_processor = DepsProcessor(...)
        deps_targets_to_resolve = deps_processor.process_deps_entry()

        This will return a list of "deps" that haven't been traversed, but are needed by
        the deps_processor entry (the target we're trying to resolve).

        This method will apply all target features to the CommandDesign ref object as
        we traverse.

        NOTE: This does not yet support conditional inclusions based on defines,
         like the old DEPS files did with pattern:
            SOME_DEFINE ?  dep_if_define_present : dep_if_define_not_present
         I would like to deprecate that in favor of 'tags'. However, likely will need
         to walk the entire DEPS.yml once to populate all args/defines, and then re-
         walk them to add/prune the correct tag based dependencies, or rely on it being
         entirely top-down.
        '''

        # DEPS.yml entries have ordered keys, and process these in-order
        # with how the <target> defined it.
        remaining_deps_list = [] # deps items we find that are not yet processed.
        for key in self.deps_entry.keys():

            # Make sure DEPS target table keys are legal:
            if key not in SUPPORTED_TARGET_TABLE_KEYS and \
               key not in self.known_eda_commands:
                error(f'Unknown target {key=} in {self.caller_info},',
                      f' must be one of opencos.deps.defaults.{SUPPORTED_TARGET_TABLE_KEYS=}',
                      f' or an eda command: {self.known_eda_commands}')

            if key == 'tags' or 'tags' in self.entry_eda_command:
                remaining_deps_list += self.process_tags()
            elif key == 'defines' or 'defines' in self.entry_eda_command:
                self.process_defines()
            elif key == 'plusargs' or 'plusargs' in self.entry_eda_command:
                self.process_plusargs()
            elif key == 'parameters' or 'parameters' in self.entry_eda_command:
                self.process_parameters()
            elif key == 'incdirs' or 'incdirs' in self.entry_eda_command:
                self.process_incdirs()
            elif key == 'top' or 'top' in self.entry_eda_command:
                self.process_top()
            elif key == 'args' or 'args' in self.entry_eda_command:
                self.process_args()
            elif key == 'commands' or 'commands' in self.entry_eda_command:
                self.process_commands()
            elif key == 'reqs' or 'reqs' in self.entry_eda_command:
                self.process_reqs()

            elif any((key == 'deps', 'deps' in self.entry_eda_command,
                      (key == 'files' and 'deps' not in self.deps_entry),
                      ('files' in self.entry_eda_command and 'deps' not in self.entry_eda_command)
                      )):
                remaining_deps_list += self.process_deps_return_discovered_deps()

            if self.command_design_ref.tool_changed_respawn:
                # Stop now, and have eda.py respawn the command.
                return []

        # We return the list of deps that still need to be resolved (['full_path/some_target', ...])
        return remaining_deps_list

    def process_tags( # pylint: disable=too-many-statements,too-many-branches,too-many-locals
            self
    ) -> list:
        '''Returns List of added deps, applies tags (dict w/ details, if any) to
        self.command_design_ref.

        Tags are only supported as a Table within a target. Current we only support:
        'args', 'replace-config-tools', 'additive-config-tools', 'with-tools', 'with-args'.
        '''

        deps_tags_enables = self.config.get('dep_tags_enables', {})
        ret_deps_added_from_tags = []

        entry_tags = {} # from yml table
        entry_tags.update(self.deps_entry.get('tags', {}))
        for tagname,value in entry_tags.items():
            debug(f'process_tags(): {tagname=} in {self.caller_info}' \
                  + f' observed: {value=}')
            assert isinstance(value, dict), \
                f'{tagname=} {value=} value must be a dict for in {self.caller_info}'
            tags_dict_to_apply = value.copy()

            for key in value.keys():
                if key not in SUPPORTED_TAG_KEYS:
                    self.error(f'{tagname=} in {self.caller_info}:',
                               f'has unsupported {key=} {SUPPORTED_TAG_KEYS=}',
                               '\n    To see detailed DEPS schema, run:',
                               f'{Colors.byellow}eda deps-schema [--verbose]',
                               error_code=status_constants.EDA_DEPS_BAD_SCHEMA_KEY)

            enable_tags_matched = False
            disable_tags_matched = False
            if tagname in self.command_design_ref.args['enable-tags']:
                # tagname was force enabled by --enable-tags=tagname.
                debug(f'process_tags(): {tagname=} in {self.caller_info=}',
                      'will be enabled, matched in --enable-tags:',
                      f'{self.command_design_ref.args["enable-tags"]}')
                enable_tags_matched = True
            if tagname in self.command_design_ref.args['disable-tags']:
                # tagname was force disabled by --disable-tags=tagname.
                debug(f'process_tags(): {tagname=} in {self.caller_info=}',
                      'will be disabled, matched in disable-tags:',
                      f'{self.command_design_ref.args["disable-tags"]}')
                disable_tags_matched = True


            apply_tag_items_tools = False
            apply_tag_items_commands = False
            apply_tag_items_with_args = False

            tool = self.args.get('tool', None)

            if disable_tags_matched or enable_tags_matched:
                # skip checking with-tools or with-args, b/c we are already
                # force matched by tagname from --enable-tags or --disable-tags.
                pass
            else:

                with_tools = dep_str2list(value.get('with-tools', []))
                if with_tools and not deps_tags_enables.get('with-tools', None):
                    with_tools = []
                    warning(f'{tagname=} in {self.caller_info}:',
                            ' skipped due to with-tools disabled.')

                with_commands = dep_str2list(value.get('with-commands', []))
                if with_commands and not deps_tags_enables.get('with-commands', None):
                    with_commands = []
                    warning(f'{tagname=} in {self.caller_info}:',
                            ' skipped due to with-commands disabled.')

                with_args = value.get('with-args', {})
                if not isinstance(with_args, dict):
                    error(f'{tagname=} in {self.caller_info}:',
                          ' with-args must be a table (dict) of key-value pairs')
                if with_args and not deps_tags_enables.get('with-args', None):
                    with_args = {}
                    warning(f'{tagname=} in {self.caller_info}:',
                            ' skipped due to with-args disabled.')

                # check with-tools?
                if not with_tools:
                    apply_tag_items_tools = True # no with-tools present
                elif tool in with_tools:
                    apply_tag_items_tools = True # with-tools present and we matched.
                else:
                    # Each item of with-tools can also be in the form
                    # {tool (str)}:{TOOL.tool_version (str)}
                    # this matches Tool.get_full_tool_and_versions()
                    if getattr(self.command_design_ref, 'get_full_tool_and_versions', None):
                        tool_full_version = self.command_design_ref.get_full_tool_and_versions()
                        if tool_full_version and tool_full_version in with_tools:
                            apply_tag_items_tools = True

                # check with-commands?
                if not with_commands:
                    apply_tag_items_commands = True # no with-commands present
                elif getattr(self.command_design_ref, 'command_name', '') in with_commands:
                    apply_tag_items_commands = True # with-commands present and we matched.

                # check with-args?
                with_args_matched_list = []
                for k,v in with_args.items():
                    with_args_matched_list.append(False)
                    if not all((apply_tag_items_tools, apply_tag_items_commands)):
                        # If we didn't previously match with-tools (if with-tools was present),
                        # or match with-commands, then we may not match the args, b/c those are
                        # command + tool dependent in the Command handling class.
                        pass
                    elif k not in self.command_design_ref.args:
                        warning(f'{tagname=} in {self.caller_info}:',
                                f'with-args key {k} is not a valid arg for {tool=}')
                    elif not isinstance(v, type(self.command_design_ref.args[k])):
                        warning(f'{tagname=} in {self.caller_info}:',
                                f' with-args table key {k} value {v} (type {type(v)}) does not',
                                f' match type in args (type {self.command_design_ref.args[k]})')
                    elif self.command_design_ref.args[k] == v:
                        # set it as matched:
                        with_args_matched_list[-1] = True
                        debug(f'{tagname=} in {self.caller_info}:',
                              f' with-args table key {k} value {v} matched')
                    else:
                        debug(f'{tagname=} in {self.caller_info}:',
                              f'with-args table key {k} value {v} did not match args value: ',
                              f'{self.command_design_ref.args[k]}')

                if not with_args_matched_list:
                    apply_tag_items_with_args = True # no with-args set
                else:
                    apply_tag_items_with_args = all(with_args_matched_list)

            # Did we match all with-tools and with-args?
            if disable_tags_matched:
                apply_tag_items = False
            elif enable_tags_matched:
                apply_tag_items = True
            else:
                apply_tag_items = all([apply_tag_items_tools, apply_tag_items_with_args,
                                       apply_tag_items_commands])

            if not apply_tag_items:
                debug(f'process_tags(): {tagname=} in {self.caller_info}',
                      f'skipped for {tool=}, {with_args=}, {with_args_matched_list=}')
            elif apply_tag_items_tools or apply_tag_items_with_args:
                debug(f'process_tags(): {tagname=} in {self.caller_info=}',
                      f'applying tags for {tool=}, {with_args=}, {with_args_matched_list=},',
                      f'{tags_dict_to_apply.keys()=}')


            if apply_tag_items:
                # We have matched something (with-tools, etc).
                # apply these in the original order of the keys:
                for key in tags_dict_to_apply.keys():

                    if key == 'defines':
                        # apply defines:
                        self.apply_defines(value.get('defines', {}))

                    if key == 'plusargs':
                        # apply plusargs:
                        self.apply_plusargs(value.get('plusargs', {}))

                    elif key == 'parameters':
                        self.apply_parameters(value.get('parameters', {}))

                    elif key == 'incdirs':
                        # apply incdirs:
                        self.apply_incdirs(value.get('incdirs', []))

                    elif key == 'args':
                        # apply args
                        args_list = dep_str2list(value.get('args', []))
                        if args_list and not deps_tags_enables.get('args', None):
                            args_list = []
                            warning(f'{tagname=} in {self.caller_info=}:',
                                    ' skipped args due to args disabled.')
                        if args_list:
                            # This will apply knowns args to the target dep:
                            debug(f'{tagname=} in {self.caller_info=}:',
                                 f'applying args for {args_list=}')
                            self.apply_args(args_list, tagname=tagname)

                    elif key == 'reqs':
                        reqs_list = deps_target_get_deps_list(entry=value,
                                                              default_key='reqs',
                                                              target_node=self.target_node,
                                                              deps_file=self.deps_file)
                        self.apply_reqs(reqs_list)

                    elif key == 'deps' or \
                         (key == 'files' and 'deps' not in tags_dict_to_apply):

                        # apply deps (includes commands, stray +define+ +incdir+)
                        # treat the same way we treat self.process_deps_return_discovered_deps
                        deps_list = deps_target_get_deps_list(entry=value,
                                                              default_key='deps',
                                                              target_node=self.target_node,
                                                              deps_file=self.deps_file)
                        ret_deps_added_from_tags += self.get_remaining_and_apply_deps(deps_list)

                # for replace-config-tools or additive-config-tools from tags, these don't need to
                # handle in order of tags keys:

                # apply replace-config-tools
                # This will replace lists (compile-waivers).
                tool_config = value.get('replace-config-tools', {}).get(tool, None)
                ref_has_tool_config = isinstance(
                    getattr(self.command_design_ref, 'tool_config', None), dict
                )
                if tool_config and (not deps_tags_enables.get('replace-config-tools', None) or \
                                    not ref_has_tool_config):
                    tool_config = None
                    warning(f'{tagname=} in {self.caller_info}:',
                            'skipped replace-config-tools b/c it is disabled or not present for',
                            'this tool and command')
                if ref_has_tool_config and tool_config and isinstance(tool_config, dict):
                    # apply it to self.tool_config:
                    info(f'{tagname=} in {self.caller_info}:',
                         f'applying replace-config-tools for {tool=}: {tool_config}')
                    eda_config.merge_config(self.command_design_ref.tool_config, tool_config)
                    # Since we altered command_design_ref.tool_config, need to call update on it:
                    self.command_design_ref.update_tool_config()
                    debug(f'{tagname=} in {self.caller_info}:',
                          'Updated {self.command_design_ref.tool_config=}')

                # apply additive-config-tools
                # This will append to lists (compile-waivers)
                tool_config = value.get('additive-config-tools', {}).get(tool, None)
                if tool_config and (not deps_tags_enables.get('additive-config-tools', None) or \
                                    not ref_has_tool_config):
                    tool_config = None
                    warning(f'{tagname=} in {self.caller_info}:',
                            ' skipped additive-config-tools b/c it is disable or not present for',
                            'this tool and command')
                if ref_has_tool_config and tool_config and isinstance(tool_config, dict):
                    # apply it to self.tool_config:
                    info(f'{tagname=} in {self.caller_info}:',
                         f'applying additive-config-tools for {tool=}: {tool_config}')
                    eda_config.merge_config(self.command_design_ref.tool_config, tool_config,
                                            additive_strategy=True)
                    # Since we altered command_design_ref.tool_config, need to call update on it:
                    self.command_design_ref.update_tool_config()
                    debug(f'{tagname=} in {self.caller_info}:',
                          f'Updated {self.command_design_ref.tool_config=}')

        return ret_deps_added_from_tags

    def _is_ok_deps_field_type(
            self, item, want_types, name: str, must_be: str = '', msg: str = '',
            error_if_item_empty: bool = False
    ) -> bool:
        '''Returns false if we don't like the the DEPS markup entry, with error message'''
        if not isinstance(item, want_types):
            if msg:
                msg = f'{msg}, '
            if not must_be:
                if want_types == (dict, list, str):
                    must_be = (
                        'markup Table/dict, Array/list, or string (space or newline separated)'
                    )
                elif want_types == (list, str):
                    must_be = 'markup Array/list, or string (space or newline separated)'
                elif want_types is dict:
                    must_be = 'markup Table/dict'
                elif want_types is list:
                    must_be = 'markup Array/list'
                elif want_types is str:
                    must_be = 'markup string (space or newline separated)'
                else:
                    assert False, f'bad {want_types=}'

            # Since all checked types here are dict, list, str, if the value is empty, allow it and
            # move on (caller shouldn't apply it)
            if not item and not error_if_item_empty:
                # item empty, no error if empty:
                return True

            self.error(f'At {self.caller_info}: {msg}"{name}:" entry must be a {must_be},',
                       f'but instead got: type={type(item)}, {name}: {item}.',
                       '\n    To see detailed DEPS schema, run:',
                       f'{Colors.byellow}eda deps-schema [--verbose]',
                       error_code=status_constants.EDA_DEPS_BAD_TYPE)
            return False

        return True


    def process_defines(self) -> None:
        '''Returns None, applies defines (dict, if any) from self.deps_entry to
        self.command_design_ref.'''

        # Defines:
        # apply command specific defines, with higher priority than the a
        # deps_entry['sim']['defines'] entry,
        # do this with dict1.update(dict2):
        entry_defines = {}
        _got_defines = self.deps_entry.get('defines', {})
        _got_cmd_defines = self.entry_eda_command.get('defines', {})

        if not self._is_ok_deps_field_type(_got_defines, (dict, list, str), name='defines'):
            return
        if not self._is_ok_deps_field_type(
                _got_cmd_defines, (dict, list, str), name='defines',
                msg=f'command entry in target ("{self.command_name}:")'
        ):
            return

        entry_defines.update(self._defines_as_dict(_got_defines))
        entry_defines.update(self._defines_as_dict(_got_cmd_defines))

        self.apply_defines(entry_defines, check=False)


    def process_plusargs(self) -> None:
        '''Returns None, applies plusargs (dict, if any) from self.deps_entry to
        self.command_design_ref.

        These work w/ the same rules as defines (no value, or value int/str)
        '''

        # Plusargs:
        # apply command specific plusargs, with higher priority than the a
        # deps_entry['sim']['plusargs'] entry,
        # do this with dict1.update(dict2):
        entry_plusargs = {}
        _got_plusargs = self.deps_entry.get('plusargs', {})
        _got_entry_eda_command_plusargs = self.entry_eda_command.get('plusargs', {})

        if not self._is_ok_deps_field_type(_got_plusargs, (dict, list, str),
                                           name='plusargs'):
            return
        if not self._is_ok_deps_field_type(
                _got_entry_eda_command_plusargs, (dict, list, str), name='plusargs',
                msg=f'command entry in target ("{self.command_name}:")'
        ):
            return

        if _got_plusargs:
            entry_plusargs.update(self._plusargs_as_dict(_got_plusargs))
        if _got_entry_eda_command_plusargs:
            entry_plusargs.update(self._plusargs_as_dict(_got_entry_eda_command_plusargs))
        assert isinstance(entry_plusargs, dict), \
            f'{entry_plusargs=} for in {self.caller_info} must be a dict'

        self.apply_plusargs(entry_plusargs, check=False)


    def process_parameters(self) -> None:
        '''Returns None, applies parameters (dict, if any) from self.deps_entry to
        self.command_design_ref.'''

        # Parameters:
        # apply command specific parameters, with higher priority than the a
        # deps_entry['sim']['parameters'] entry,
        # do this with dict1.update(dict2):
        entry_parameters = {}
        _got_params = self.deps_entry.get('parameters', {})
        _got_entry_eda_command_params = self.entry_eda_command.get('parameters', {})

        if not self._is_ok_deps_field_type(_got_params, (dict, list, str),
                                           name='parameters'):
            return
        if not self._is_ok_deps_field_type(
                _got_entry_eda_command_params, (dict, list, str), name='parameters',
                msg=f'command entry in target("{self.command_name}:")'
        ):
            return

        if _got_params:
            entry_parameters.update(self._parameters_as_dict(_got_params))
        if _got_entry_eda_command_params:
            entry_parameters.update(self._parameters_as_dict(_got_entry_eda_command_params))
        assert isinstance(entry_parameters, dict), \
            f'{entry_parameters=} for in {self.caller_info} must be a dict'

        self.apply_parameters(entry_parameters, check=False)


    def process_incdirs(self) -> None:
        '''Returns None, applies incdirs (dict, if any) from self.deps_entry to
        self.command_design_ref.'''

        entry_incdirs = []
        _got_incdirs = self.entry_eda_command.get('incdirs', [])
        _got_entry_eda_command_incdirs = self.deps_entry.get('incdirs', [])
        if not self._is_ok_deps_field_type(
                _got_incdirs, want_types=(list, str), name='incdirs',
        ):
            return
        if not self._is_ok_deps_field_type(
                _got_entry_eda_command_incdirs, want_types=(list, str),
                name='incdirs', msg=f'command entry in target ("{self.command_name}:")'
        ):
            return

        # apply command specific incdirs, higher in the incdir list:
        if _got_incdirs:
            entry_incdirs = dep_str2list(_got_incdirs)
        if _got_entry_eda_command_incdirs:
            entry_incdirs += dep_str2list(_got_entry_eda_command_incdirs)
        assert isinstance(entry_incdirs, list), \
            f'{entry_incdirs=} for in {self.caller_info} must be a list'
        self.apply_incdirs(entry_incdirs)

    def process_top(self) -> None:
        '''Returns None, applies top (str, if any) from self.deps_entry to
        self.command_design_ref.'''

        if self.args['top'] != '':
            return # already set

        _got_top = self.deps_entry.get('top', '')            # target.top
        _got_cmd_top = self.entry_eda_command.get('top', '') # target.sim.top for example
        if not self._is_ok_deps_field_type(_got_top, str, name='top'):
            return
        if not self._is_ok_deps_field_type(_got_cmd_top, str, name='top',
                                          msg=f'command entry in target ("{self.command_name}:")'):
            return

        # For 'top', we overwrite it if not yet set.
        # the command specific 'top' has higher priority.
        entry_top = _got_cmd_top
        if entry_top == '':
            entry_top = _got_top

        if entry_top != '':
            if self.args['top'] == '':
                # overwrite only if unset - we don't want other deps overriding the topmost
                # target's setting for 'top'.
                self.set_arg('top', str(entry_top))

    def process_args(self) -> None:
        '''Returns None, applies args (list or str, if any) from self.deps_entry to
        self.command_design_ref.'''

        # for 'args', process each. command specific args take higher priority that target args.
        # run_argparser_on_list: uses argparse, which takes precedence on the last arg that is set,
        # so put the command specific args last.
        # Note that if an arg is already set, we do NOT update it
        _got_args = self.deps_entry.get('args', [])
        _got_cmd_args = self.entry_eda_command.get('args', [])
        if not self._is_ok_deps_field_type(
                _got_args, want_types=(list, str), name='args'
        ):
            return
        if not self._is_ok_deps_field_type(
                _got_cmd_args, want_types=(list, str), name='args',
                msg=f'command entry in target ("{self.command_name}:")'
        ):
            return

        args_list = []
        if _got_args:
            args_list = dep_str2list(_got_args)
        if _got_cmd_args:
            args_list += dep_str2list(_got_cmd_args)

        # for args_list, re-parse these args to apply them to self.args.
        if not args_list:
            return

        debug(f'in {self.caller_info}: {args_list=}')
        self.apply_args(args_list)

        # TODO(drew): Currently, I can't support changing the 'config' via an arg encountered in
        # DEPS.yml. This is prevented b/c --config-yml appears as a modifed arg no matter what
        # (and we don't let DEPS.yml override modifed args, otherwise a target would override the
        # user command line).


    def get_commands( # pylint: disable=dangerous-default-value
            self, commands: list = [], dep: str = ''
    ) -> (list, list):
        '''Returns tuple of (shell_commands_list, work_dir_add_srcs_list).

        Does not have side effects on self.command_design_ref.
        '''

        default_ret = [], []

        if not commands:
            # if we weren't passed commands, then get them from our target (self.deps_entry)
            commands = self.deps_entry.get('commands', [])

        if not self._is_ok_deps_field_type(commands, list, name='commands'):
            return default_ret

        if not commands: # No commands in this target
            return default_ret

        debug(f"Got {self.deps_entry=} for in {self.caller_info}, has {commands=}")
        shell_commands_list = [] # list of dicts
        work_dir_add_srcs_list = [] # list of dicts

        if not dep:
            # if we weren't passed a dep, then use our target_node (str key for our self.deps_entry)
            dep = self.target_node

        # Run handler for this to convert to shell commands in self.command_design_ref
        shell_commands_list, work_dir_add_srcs_list = deps_commands_handler(
            config=self.command_design_ref.config,
            eda_args=self.command_design_ref.args,
            dep=dep,
            deps_file=self.deps_file,
            target_node=self.target_node,
            target_path=self.target_path,
            commands=commands
        )

        return shell_commands_list, work_dir_add_srcs_list

    def process_commands( # pylint: disable=dangerous-default-value
            self, commands: list = [], dep: str = ''
    ) -> None:
        '''Returns None, handles commands (shell, etc) in the target that aren' in the 'deps' list.

        Applies these to self.command_design_ref.

        You can optionally call this with a commands list and a single dep, which we support for
        commands lists that exist within the 'deps' entry of a target.
        '''

        shell_commands_list, work_dir_add_srcs_list = self.get_commands(commands=commands, dep=dep)

        if shell_commands_list and \
           not self.is_command_design:
            warning(f'Not applying shell commands from {self.caller_info}, not supported',
                    'for this tool and command')
            return

        # add these commands lists to self.command_design_ref:
        # Process all shell_commands_list:
        # This will track each shell command with its target_node and target_path
        self.command_design_ref.append_shell_commands( cmds=shell_commands_list )
        # Process all work_dir_add_srcs_list:
        # This will track each added filename with its target_node and target_path
        self.command_design_ref.append_work_dir_add_srcs( add_srcs=work_dir_add_srcs_list,
                                                          caller_info=self.caller_info )


    def process_reqs(self) -> None:
        '''Process any 'reqs:' table in a DEPS markup entry'''
        reqs_list = deps_target_get_deps_list(entry=self.deps_entry,
                                              default_key='reqs',
                                              target_node=self.target_node,
                                              deps_file=self.deps_file)
        if not self._is_ok_deps_field_type(reqs_list, want_types=(list, str), name='reqs'):
            return
        if reqs_list:
            self.apply_reqs(reqs_list)


    def process_deps_return_discovered_deps(self) -> list:
        '''Returns list of deps targets to continue processing,

        -- iterates through 'deps' for this target (self.deps_entry['deps'])
           -- note this will also append 'files' if that table key + list/str exists.
        -- applies to self.command_design_ref
        '''

        # Get the list of deps from this entry (entry is a target in our DEPS.yml):
        deps = deps_target_get_deps_list(
            self.deps_entry,
            target_node=self.target_node,
            deps_file=self.deps_file
        )
        if not self._is_ok_deps_field_type(deps, want_types=(dict, list, str), name='deps'):
            return []
        return self.get_remaining_and_apply_deps(deps)

    def get_remaining_and_apply_deps(
            self, deps: list
    ) -> list:
        '''Given a list of deps, process what is supported in a "deps:" table in DEPS
        markup file.'''

        deps_targets_to_resolve = []

        # Process deps (list)
        for dep in deps:
            deps_targets_to_resolve.extend(
                self._get_remaining_and_apply_single_dep(dep)
            )
        return deps_targets_to_resolve


    def _get_remaining_and_apply_single_dep( # pylint: disable=too-many-branches
            self, dep: str
    ) -> list:
        '''Given a single dep, process is and return targets/files that need resolving'''

        deps_targets_to_resolve = []
        typ = type(dep)
        if typ not in SUPPORTED_DEP_KEYS_BY_TYPE:
            self.error(f'{self.target_node=} {dep=} in {self.deps_file=}:' \
                       + f'has unsupported {type(dep)=} {SUPPORTED_DEP_KEYS_BY_TYPE=}',
                       '\n    To see detailed DEPS schema, run:',
                       f'{Colors.byellow}eda deps-schema [--verbose]',
                       error_code=status_constants.EDA_DEPS_BAD_TYPE)

        for supported_values in SUPPORTED_DEP_KEYS_BY_TYPE.values():
            if '*' in supported_values:
                continue
            if typ in (dict, list) and any(k not in supported_values for k in dep):
                self.error(
                    f'{self.target_node=} {dep=} in {self.deps_file=}: has dict-key or',
                    f'list-item not in {SUPPORTED_DEP_KEYS_BY_TYPE[typ]=}',
                    '\n    To see detailed DEPS schema, run:',
                    f'{Colors.byellow}eda deps-schema [--verbose]',
                    error_code=status_constants.EDA_DEPS_BAD_TYPE
                )

        # In-line commands in the deps list, in case the results need to be in strict file
        # order for other deps
        if isinstance(dep, dict) and 'commands' in dep:

            commands = dep['commands']
            debug(f"Got commands {dep=} for in {self.caller_info}, {commands=}")

            assert isinstance(commands, list), \
                f'dep commands must be a list: {dep=} in {self.caller_info}'

            # For this, we need to get the returned commands (to keep strict order w/ other
            # deps)
            command_tuple = self.get_commands( commands=commands, dep=dep )
            # TODO(drew): it might be cleaner to return a dict instead of list, b/c those
            # are also ordered and we can pass type information, something like:
            deps_targets_to_resolve.append(command_tuple)


        elif isinstance(dep, str) and \
             any(dep.startswith(x) for x in ['+define+', '+incdir+']) and \
             self.is_command_design:
            # Note: we still support +define+ and +incdir in the deps list.
            # check for compile-time Verilog style plusarg, which are supported under targets
            # These are not run-time Verilog style plusargs comsumable from within the .sv:
            debug(f"Got plusarg (define, incdir) {dep=} for {self.caller_info}")
            self.command_design_ref.process_plusarg(plusarg=dep, pwd=self.target_path)

        else:
            # If we made it this far, dep better be a str type.
            assert isinstance(dep, str), f'{dep=} {type(dep)=} must be str'
            dep_path = self.correct_a_deps_target(target=dep, deps_dir=self.target_path)
            debug(f"Got dep {dep_path=} for in {self.caller_info}")

            if self.is_command_design and \
               dep_path in self.command_design_ref.targets_dict or \
               dep_path in deps_targets_to_resolve:
                debug(f" - already processed ({dep_path}), skipping")
            else:
                # This is where we support files/deps file wildcards via glob syntax.
                # If glob found none, we fall back to a single-file or target
                glob_added_from_dep_path = False
                if any(x in dep_path for x in ('*', '?', '[')):
                    try:
                        glob_list = glob.glob(dep_path)
                        for fpath in glob_list:
                            file_exists, _, _ = files.get_source_file(fpath)
                            if file_exists and fpath not in deps_targets_to_resolve:
                                debug(f" - raw file ({fpath}), from glob ({dep_path}) adding",
                                      "to return list...")
                                deps_targets_to_resolve.append(fpath) # append
                                glob_added_from_dep_path = True
                        if glob_list and not glob_added_from_dep_path:
                            self.error(f'No files were expanded from glob: {dep_path} in',
                                       f'{self.caller_info}',
                                       error_code=status_constants.EDA_DEPS_TARGET_NOT_FOUND)
                    except Exception as e:
                        self.error(f'Unable to add files via glob {dep_path}, in',
                                   f'{self.caller_info}exception {e}',
                                   error_code=status_constants.EDA_DEPS_TARGET_NOT_FOUND)

                if not glob_added_from_dep_path:
                    file_exists, _, _ = files.get_source_file(dep_path)
                    if file_exists:
                        debug(f" - raw file ({dep_path}), adding to return list...")
                        deps_targets_to_resolve.append(dep_path) # append, keeping file order.
                    else:
                        debug(f" - a target non-file ({dep_path}) needs to be resolved,",
                              "adding to return list...")
                        deps_targets_to_resolve.append(dep_path) # append, keeping file order.

        # We return the list of deps or files that still need to be resolved
        # (['full_path/some_target', ...])
        # items in this list are either:
        #  -- string (dep or file)
        #  -- tuple (unprocessed commands, in form: (shell_commands_list, work_dir_add_srcs_list))
        return deps_targets_to_resolve


    def correct_a_deps_target(self, target: str, deps_dir: str) -> str:
        '''Give a target/file in a deps: list, return a patched version

        - $VAR replacment
        - relative to current DEPS file dir (or not if it was abspath)
        '''
        if self.config['deps_expandvars_enable']:
            target = os.path.expandvars(target)
        if not os.path.isabs(target):
            target = os.path.join(deps_dir, target)
        return target



'''Handler for: eda waves --tool=surfer'''

import os

from opencos.commands.waves import CommandWaves
from opencos.tools.vaporview import ToolVaporview

class ToolSurfer(ToolVaporview):
    '''Run Surfer the same way we'd run Vaporview - via VsCode extension

    Using ToolVaporview to reconfigure it, the extesion is selected from
    eda's CONFIG-YML config.tools.surfer (has extension name)
    '''

    _TOOL = 'surfer'
    _EXE = 'code'
    _URL = 'https://surfer-project.org/'


class CommandWavesSurfer(CommandWaves, ToolSurfer):
    '''Command handler for: eda waves --tool=surfer

    This is registerd in eda's CONFIG-YML config.tools.surfer.handlers.waves
    '''

    def __init__(self, config: dict):
        CommandWaves.__init__(self, config=config)
        ToolSurfer.__init__(self, config=self.config)

    def do_it(self) -> None:
        command_list = [
            self.waveviewer_exe, '--disable-workspace-trust', '-r', self.wave_file
        ]
        self.exec(os.path.dirname(self.wave_file), command_list)



''' private helper package for `eda --version`'''

import sys

# pylint: disable=invalid-name

NAME = 'opencos-eda'

# Use standard library metadata module starting Python 3.8
if sys.version_info >= (3, 8):

    from importlib import metadata
    try:
        ver = metadata.version(NAME)
    except metadata.PackageNotFoundError:
        # Handle case where the package is not installed (e.g., running from source checkout)
        ver = "0.0.0"

else:
    # This package only supports >= 3.8, so not doing the importlib_metadata method.
    ver = "unknown" # Or raise an error, or handle differently


VERSION = ver



'''
Macos utils for simple things in other OSes, such as: finding anything in
/Applications/
'''

import os
import plistlib
from opencos import eda_config
from opencos.utils import subprocess_helpers

IS_MACOS = subprocess_helpers.IS_MACOS

def get_macos_tool_info( # pylint: disable=dangerous-default-value
        app_name: str, prefer_exes: list = []
) -> dict:
    """
    Resolves the actual binary path and version from a macOS .app bundle.
    app_name: e.g., 'Visual Studio Code.app'

    Will look in:
      -- $HOME/Applications/
      -- /Applications/
    """
    ret = {}
    if not IS_MACOS:
        return ret

    if not app_name.endswith('.app'):
        app_name += '.app'

    # Try $HOME/Applications first.
    home = os.environ.get('HOME', '')
    if home:
        base_path = os.path.join(home, '/Applications', app_name)

    if not os.path.exists(base_path):
        # If not in $HOME/Applications, try /Applications:
        base_path = os.path.join('/Applications', app_name)

    plist_path = os.path.join(base_path, 'Contents/Info.plist')

    ret.update({
        'base_path': {
            'value': base_path,
            'exists': os.path.exists(base_path)
        },
        'plist_path': {
            'value': plist_path,
            'exists': os.path.exists(plist_path)
        }
    })

    if not os.path.exists(plist_path):
        return ret

    with open(plist_path, 'rb') as f:
        info = plistlib.load(f)

    # 1. Resolve Executable (Replaces shutil.which)
    exe_name = info.get('CFBundleExecutable')
    exe_path = os.path.join(base_path, 'Contents/MacOS', exe_name)

    # 2. Resolve Version (Standard macOS way)
    version = info.get('CFBundleShortVersionString', 'Unknown')

    # 3. Custom Metadata (Your product.json)
    # Usually in Resources, but could be in Contents
    product_json_path = os.path.join(base_path, 'Contents/Resources/product.json')
    if not os.path.exists(product_json_path):
        # we can try Contents/Resources/app/product.json:
        product_json_path = os.path.join(
            base_path, 'Contents/Resources/app/product.json')

    alt_bin_path = os.path.join(base_path, 'Contents/Resources/bin')
    if not os.path.isdir(alt_bin_path):
        alt_bin_path = os.path.join(base_path, 'Contents/Resources/app/bin')

    ret.update({
        'product_json_path': {
            'value': product_json_path,
            'exists': os.path.exists(product_json_path)
        },
        'alt_bin_path': {
            'value': alt_bin_path,
            'exists': os.path.isdir(alt_bin_path)
        }
    })

    if alt_bin_path:
        for prefer_exe in reversed(prefer_exes):
            prefer_exe_path = os.path.join(alt_bin_path, os.path.split(prefer_exe)[1])
            if os.path.exists(prefer_exe_path):
                exe_path = prefer_exe_path

    ret.update({
        "bin": exe_path,
        "version": version,
        "metadata": product_json_path if os.path.exists(product_json_path) else ''
    })
    return ret


def get_tool_exe_from_config(config: dict, tool: str) -> str:
    '''Returns the macos /Applications/ exe path using 'tool'

    Also will look at $HOME/Applications (using get_macos_tool_info(str))

    If no macos, reverts to the first exe in config.
    tool -- this is the opencos-eda tool name, such as 'verilator',
    from the --tool=TOOL arg, or auto-found.
    '''

    tool_cfg = config.get('tools', {}).get(tool, {})
    exe_list = eda_config.tool_get_exe_list(tool=tool, config=config)
    first_exe = exe_list[0]

    if IS_MACOS:
        app_name = tool_cfg.get('macos-app-name', '')
        if not app_name:
            app_name = tool_cfg.get('app-name', '')
        if not app_name:
            app_name = tool_cfg.get('windows-app-name', '')
        prefer_exes = tool_cfg.get('macos-prefer-exes', [])
        if exe_list:
            prefer_exes = exe_list + prefer_exes
        if app_name:
            data = get_macos_tool_info(app_name, prefer_exes=prefer_exes)
            if e := data.get('bin', ''):
                first_exe = e

    return first_exe



#!/usr/bin/env python3

'''
Helper pymodule used by eda_deps_bash_completion.bash, extracts valid
targets from DEPS files
'''

import sys
import os
from pathlib import Path

from opencos.deps.deps_file import get_all_targets
from opencos.utils.str_helpers import print_columns_manual

PATH_LPREFIX = str(Path('.')) + os.path.sep


def get_path_and_pattern(partial_path: str = '', base_path=str(Path('.'))) -> (str, str):
    '''Returns tuple of (partial_path, partial_target or filter)'''

    partial_target = ''
    if not partial_path or Path(partial_path) == Path('.'):
        partial_path = PATH_LPREFIX

    # if the partial path is not an existing file or dir, then treat it as
    # a partial and split it so partial_path/partial_target can be used later.
    if not base_path or Path(base_path) == Path('.'):
        if not os.path.exists(partial_path):
            partial_path, partial_target = os.path.split(partial_path)
    else:
        if not os.path.exists(os.path.join(base_path, partial_path)):
            partial_path, partial_target = os.path.split(partial_path)

    # if we have no partial path, use compatible ./
    if not partial_path:
        partial_path = PATH_LPREFIX

    return partial_path, partial_target


def get_targets(partial_paths: list, base_path=str(Path('.'))) -> list:
    '''Returns a list of DEPS keys into pretty columns, using arg

    partial_path as a string filter for target completions.
    '''

    targets_set = set()
    if not partial_paths:
        partial_paths = [PATH_LPREFIX] # run on current directory.

    for partial_path in partial_paths:
        partial_path, partial_target = get_path_and_pattern(
            partial_path=partial_path, base_path=base_path
        )
        try:
            keys = get_all_targets(
                dirs=[partial_path],
                base_path=base_path,
                filter_str=partial_target,
                error_on_empty_return=False,
                lstrip_path=True
            )
        except Exception:
            keys = []
        for key in keys:
            targets_set.add(key)

    return list(targets_set)


def run(partial_paths: list, base_path=str(Path('.'))) -> None:
    '''Returns None, prints DEPS keys into pretty columns, using arg

    partial_path as a string filter for target completions.
    '''

    data = get_targets(partial_paths=partial_paths, base_path=base_path)
    data.sort()
    print_columns_manual(data=data, num_columns=4, auto_columns=True)


def main() -> None:
    '''Returns None, prints DEPS keys into pretty columns, uses sys.argv[1] for args:

    sys.args[1] -- optional partial path for target completions
    '''

    if len(sys.argv) > 1:
        partial_path = sys.argv[1]
    else:
        partial_path = ''
    run(partial_paths=[partial_path])


if __name__ == "__main__":
    main()



"""
Utility functions using slang for SystemVerilog analysis.

This is to avoid incompatiblity with pyslang (conflicts with some other
EDA tool installs) and it generally lagging behind slang releases.
"""

import json
import os
from pathlib import Path
import subprocess

from opencos import util
from opencos.commands import sim
from opencos.files import safe_shutil_which
from opencos.utils.str_helpers import sanitize_defines_for_sh


INSTALL_HELPER_STR = '''
Install "slang" with:
       git clone https://github.com/MikePopoloski/slang
       cd slang
       cmake -B build
       cmake --build build -j
       cd build
       sudo make install
'''

def warn_slang_not_available(extra_text: str = '') -> None:
    '''Warning if slang isn't available on this system'''
    if 'slang_not_installed' in util.AlreadyMessage.warn_names:
        return
    util.AlreadyMessage.warn_names.add('slang_not_installed')
    util.warning(f'External tool "slang" is required {extra_text}')
    for line in INSTALL_HELPER_STR.split('\n'):
        util.warning(line)


class SlangAvailable:
    '''
    Helper to report availability of 'slang', via PATH or caller supplied EXE

    Used by opencos.tools.slang.ToolSlang
    '''

    available = None
    version: str = ''
    exe: str = ''
    base_path: str = ''

    def init(self, exe: str = '', quiet: bool = False) -> None:
        '''Sets up class members, outside users can get at this via:
            > slang_helpers.available.available

        Does not run if it has already been run and exe arg is empty, which allows
        slang.ToolSlang.get_versions to re-run init() if using a specific exe path.
        '''
        if self.available is not None:
            if not exe:
                return # already set, on default exe
        if exe:
            assert os.path.split(exe)[-1] == 'slang', f'Bad {exe=} arg to init(..)'
        else:
            exe = 'slang'

        # members not set, or set but someone BYO'd exe
        self.exe = safe_shutil_which(exe)
        self.base_path, _ = os.path.split(self.exe)
        self.available = bool(self.exe)
        if self.available:
            self.version = self._get_version(self.exe, quiet=quiet)

    def _get_version(self, exe: str = '', quiet: bool = False) -> str:
        '''Gets the version from: slang --version, called by init()'''

        if not exe:
            exe = self.exe
        if not exe:
            return ''
        version_ret = subprocess.run([exe, '--version'], capture_output=True, check=False)
        stdout = version_ret.stdout.decode('utf-8', errors='replace')
        words = stdout.split() # slang version 8.0.6+b4a74b00
        if len(words) < 3:
            if not quiet:
                util.warn_once(f'{exe} --version: returned unexpected string {version_ret=}',
                               warn_name=f'slang_version__{exe}')
            return ''
        version = words[2]
        left, _ = version.split('+')
        ver_list = left.split('.')
        if len(ver_list) != 3:
            if not quiet:
                util.warn_once(f'{exe} --version: returned unexpected string',
                               f'{version_ret=} {version=}',
                               warn_name=f'slang_version__{exe}')
            return ''
        return left


# Have a global version holding available, exe, version:
slang_available = SlangAvailable()

def check_slang_available() -> bool:
    '''Returns True is slang is available for use'''
    if slang_available.available is None:
        slang_available.init()
    return slang_available.available


def slang_read_ast_json(filepath: str) -> dict:
    '''Get data from AST'''
    ret = {}
    with open(filepath, 'r', encoding='utf-8', errors='ignore') as f:
        ret = json.load(f)
    return ret


class SlangAst:
    '''Methods for getting data from an existing AST slang.json file, or generating

    a new one via subprocess'''

    hier_sep: str = '/'
    uvm_home: str = os.environ.get('UVM_HOME', '') # best effort path to location of uvm_pkg.sv
    needs_uvm: bool = False # default
    handler_obj = None
    threads: int = 1 # we try to get from the handler_obj if possible

    slang_command_args = [
        '--timescale=1ns/1ns', '--single-unit', '--allow-dup-initial-drivers',
        '--ast-json-source-info', '--ast-json-detailed-types',
        '-Wno-explicit-static', '-Wno-missing-top',
    ]
    slang_json_fname = 'slang.json'
    slang_sh_fname = 'run_slang_for_ast.sh'
    work_dir = ''
    timeout: int = 180 # seconds, 3 min for slang to run upper-bound? Usually its 1sec or less

    def __init__(self, handler_obj: object,
                 filepath: str, hier_sep: str = '', uvm_home: str = ''):
        self.handler_obj = handler_obj

        if hier_sep:
            self.hier_sep = hier_sep

        self.ast_data = None
        self.root_node = None
        self.all_items = [] # clear these

        # deal with UVM:
        self.set_uvm_and_uvm_home(needs_uvm=True, uvm_home=uvm_home)

        # load AST if provided:
        self.init_ast_filepath(filepath)

        # inherit the max threads from the handler obj:
        self.threads = getattr(handler_obj, 'threads',
                               getattr(handler_obj, 'args', {}).get('threads', 1))


    def init_ast_filepath(self, filepath: str) -> None:
        '''Call if you set your constructor with filepath='' or None,

        sets self.ast_data, self.root_node; clears self.all_items.
        '''
        if not filepath:
            return

        self.ast_data = slang_read_ast_json(filepath)
        self.root_node = self.ast_data.get('design', {})
        self.all_items = []


    def set_uvm_and_uvm_home(self, needs_uvm: bool, uvm_home: str) -> None:
        '''Sets self.needs_uvm, overwrites self.uvm_home'''
        if not needs_uvm or not uvm_home:
            # do not update, but clear self.needs_uvm
            self.needs_uvm = False
            return


        # deal with Windows:
        p = Path(uvm_home)
        if p.is_dir() and (p / 'uvm_pkg.sv').is_file():
            self.needs_uvm = True
            self.uvm_home = str(p.resolve())
            return

        self.needs_uvm = False
        self.uvm_home = ''
        util.error(f'Attempt to set SlangAst uvm_home={uvm_home} failed, either not',
                   'a directory, or file uvm_pkg.sv not present')


    def run_slang_for_ast_auto(self) -> None:
        '''Wrapper for run_slang_for_ast(...)

        Will set self.handler_obj.has_pre_compile_slang_sh = self.slang_sh_fname,
        if that attr exists in self.handler_obj
        '''
        assert self.handler_obj, 'Need a handler_obj set to run this method'
        self.run_slang_for_ast(
            defines=self.handler_obj.defines,
            incdirs=self.handler_obj.incdirs,
            parameters=self.handler_obj.parameters,
            files=(self.handler_obj.files_sv + self.handler_obj.files_v),
            work_dir=self.handler_obj.args.get('work-dir', '')
        )


    def run_slang_for_ast( # pylint: disable=too-many-branches
            self,
            defines: dict,
            incdirs: list,
            parameters: dict,
            files: list,
            work_dir: str = ''
    ) -> None:
        '''Attemps to run slang to get an AST slang.json

        Will set self.handler_obj.has_pre_compile_slang_sh = self.slang_sh_fname,
        if that attr exists in self.handler_obj
        '''

        slang_available.init()
        if not slang_available.available or not slang_available.exe:
            if t := getattr(self.handler_obj, '_TOOL', ''):
                tool_str = f'by --tool={t}'
            else:
                tool_str = ''
            warn_slang_not_available(f'for detailed coverage {tool_str}')
            return


        command_list = [
            slang_available.exe, # from pymodule global
        ] + self.slang_command_args + [
            '--ast-json', self.slang_json_fname
        ]

        if not any((a.startswith('-j') or a.startswith('--threads')) for a in command_list):
            command_list.extend( ['-j', str(self.threads) ] )

        # incdirs:
        if self.needs_uvm and self.uvm_home:
            command_list.extend(['--include-directory', self.uvm_home])
        for value in incdirs:
            command_list.extend(['--include-directory', value])

        # defines (caller may choose to send no defines, and instead use
        # CommandSim.create_ext_defines_sv(obj)
        for k,v in defines.items():
            command_list.append( '--define-macro' )
            if v is None:
                command_list.append(k)
            else:
                # Generally we should only support int and str python types passed as
                # --define-macro {k}={v}
                command_list.append(f'{k}={sanitize_defines_for_sh(v)}')

        # parameters:
        command_list.extend(
            sim.parameters_dict_get_command_list(params=parameters, arg_prefix='-G')
        )

        # Note - As of slang <= 10.0.30, do not be tempted to add uvm_pkg.sv or uvm.sv,
        # even if you have an know their whereabouts (self.uvm_home and self.needs_uvm).
        # Slang can parse them, but will not build you an AST slang.json, which is what
        # we ultimately need.

        # files:
        command_list.extend(files)

        # run from work-dir
        assert hasattr(self.handler_obj, 'exec'), f'{self.handler_obj=} does not have exec method'

        # This command needs to make it into all.sh in the work-dir, otherwise
        # the only way to re-run is through "eda COMMAND ..."
        if self.slang_sh_fname:
            util.write_shell_command_file(
                dirpath=work_dir, filename=self.slang_sh_fname,
                command_lists=[command_list], line_breaks=False
            )
            if os.path.exists(os.path.join(work_dir, self.slang_sh_fname)) and \
               hasattr(self.handler_obj, 'has_pre_compile_slang_sh'):
                self.handler_obj.has_pre_compile_slang_sh = self.slang_sh_fname

        # Using self.handler_obj.exec timeouts, we may want to use a default value
        # if the handler_obj doesn't have a subprocess-timeout arg set.
        _, stdout, rc = self.handler_obj.exec(
            work_dir=work_dir,
            command_list=command_list,
            background=False,
            stop_on_error=False,
            quiet=False,
            tee_fpath='slang.ast.log',
            shell=False,
            ignore_error=True,
            timeout=self.timeout,
            timeout_str=f'SlangAst internal timeout ({self.timeout} seconds)'
        )
        slang_json_fpath = os.path.join(work_dir, self.slang_json_fname)
        if os.path.isfile(slang_json_fpath):
            util.info(f'slang_helpers.SlangAst.run_slang_for_ast({work_dir=}, ...):',
                      f'wrote slang AST to: {slang_json_fpath}')
            self.init_ast_filepath(filepath=slang_json_fpath)
            return

        util.warning(f'slang_helpers.SlangAst.run_slang_for_ast({work_dir=}, ...)',
                     f'unable to generate {self.slang_json_fname}')

        if rc != 0:
            util.warning(f'slang_helpers.SlangAst.run_slang_for_ast({work_dir=}, ...)',
                         f'returned non-zero return code: {rc}')
        elif 'Build failed:' in stdout:
            util.warning(f'slang_helpers.SlangAst.run_slang_for_ast({work_dir=}, ...)',
                         'build was unsuccessful')


    @staticmethod
    def static_covergroups_extract_details(name: str, member, only_covergroups: list):
        """Specialized helper to dive into CovergroupType / CovergroupBody."""
        if not name:
            cvg_name = member.get("name", "unnamed_cg")
        else:
            cvg_name = name
        points = {}

        collect_enabled = only_covergroups is not None

        if not collect_enabled:
            return {}

        collect_all = collect_enabled and not only_covergroups # empty list


        # Covergroups have a 'CovergroupBody' in their members
        for m in member.get("members", []):
            if m.get("kind") == "CovergroupBody":
                # The Body contains the actual Coverpoints
                for body_m in m.get("members", []):
                    if body_m.get("kind") == "Coverpoint":
                        cp_name = body_m.get("name", "unnamed_cp")

                        collect = bool(
                            collect_all or
                            cvg_name in only_covergroups or
                            cp_name in only_covergroups
                        )

                        # Extract Bins if you need them for explicit coverage hits
                        bins = [b.get("name") for b in body_m.get("members", [])
                                if b.get("kind") == "CoverageBin"]

                        if collect:
                            points[cp_name] = bins

        if points or \
           not only_covergroups or \
           cvg_name in only_covergroups:
            return {"name": cvg_name, "points": points}

        return {}

    def get_items_all(self) -> list:
        '''Gets all items from get_items(..) starting at root'''
        self.all_items = self.get_items(node=self.root_node)
        return self.all_items


    # pylint: disable=unknown-option-value
    # pylint: disable=dangerous-default-value
    # pylint: disable=too-many-locals,too-many-branches,too-many-statements
    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-positional-arguments
    def get_items(
            self,
            node: dict = {},  # current node in dict AST.
            path: str = "",  # current path in the hier, "" is start at root.
            traversed_modules: list = [], # list of all modules up to this point (incl itself)
            only_top_module: str = '',    # if set, value must be in traversed_modules
            only_module_types: list = [], # if empty, use all
            only_files: list = [],        # leaf file names, if empty use all.
            only_paths: list = [],        # paths if current path starts with them
            only_ports: list = [],        # if empty, use all: 'In', 'Out', 'Inout']
            only_signals: list = [],      # if empty, use all
            only_covergroups: list = [],
            max_depth: int = 99
    ) -> list:
        """
        Recursively walks the 'design' hierarchy in a slang JSON file.
        """

        collect_ports_enabled = only_ports is not None
        collect_signals_enabled = only_signals is not None
        collect_covergroups_enabled = only_covergroups is not None


        if not node:
            node = self.root_node
        if not node:
            return []

        # 1. Identify if this node is an Instance or GenerateBlock
        #    but collect some details first, if they exist
        node_kind = node.get("kind", "")
        members = node.get("members", [])
        body = node.get("body", {})
        if body and isinstance(body, dict):
            if members_alt := body.get("members", []):
                members = members_alt
        else:
            # TODO(drew): shortcut here, but if body is a str,
            # it is a space separatated "(addr) (module-type)"
            # you would have to go find the addr previously in the
            # ast... skipping this for now.
            body = {}

        inst_name = node.get("name", "unnamed")
        if path:
            current_path = f"{path}{self.hier_sep}{inst_name}"
        else:
            current_path = inst_name

        # Usually will traverse $root, '', all packages, then TOP/, ..
        util.debug(f'SlangAst.get_items -- start {current_path=} {max_depth=}')

        if node_kind in ("Instance", "GenerateBlock", "Package"):

            # v10.0 uses 'source_file' directly
            source_file = node.get("source_file", "unknown")

            # 2. Extract Ports and find Child Instances from the body
            if body.get('kind', '') in ('InstanceBody', 'PackageBody'):
                module_type = body.get('name', 'unknown')
                source_file = body.get("source_file", "unknown")
            else:
                module_type = ''

            ports = []
            signals = []
            covergroups = []
            child_results = []

            collect = True
            if all((only_module_types, module_type, module_type not in only_module_types)):
                collect = False
                util.debug(f'SlangAst.get_items -- not collecting at {current_path},',
                           f'{module_type=} not in {only_module_types=}')
            elif only_files and os.path.split(source_file)[-1] not in only_files:
                collect = False
                util.debug(f'SlangAst.get_items -- not collecting at {current_path},',
                           f'file {os.path.split(source_file)[-1]} not in {only_files=})')
            elif only_paths and not any(current_path.startswith(p) for p in only_paths):
                # TODO(drew): it might be worth skipping the only_top_module path is that
                # is set, that way we can match on u_dut/ without needing testbench_name/u_dut/
                collect = False
                util.debug(f'SlangAst.get_items -- not collecting at {current_path},',
                           f'not in {only_paths=})')
            elif only_top_module and (only_top_module not in traversed_modules + [module_type]):
                collect = False
                util.debug(f'SlangAst.get_items -- not collecting at {current_path},',
                           f'{only_top_module=} not in traversed modules')

            for m in members:
                kind = m.get("kind", "")
                name = m.get("name", "")


                # Collect ports
                if collect and collect_ports_enabled and kind == "Port":
                    direction = m.get("direction", "")
                    real_name = m.get("internalSymbol", "").split()[-1]
                    if real_name and (not only_ports or direction in only_ports):
                        ports.append({
                            real_name: {
                                "direction": m.get("direction", "")
                            }
                        })

                # Collect signals
                elif collect and collect_signals_enabled and kind in ("Net", "Variable"):
                    if name and (not only_signals or name in only_signals):
                        signals.append(name)

                # Collect Coverage (Covergroups/Points)
                elif collect and collect_covergroups_enabled and kind == "CovergroupType":
                    # Covergroups often have nested Coverpoints in their own body, call helper
                    covergroups.append(SlangAst.static_covergroups_extract_details(
                            name=name, member=m, only_covergroups=only_covergroups
                    ))

                # 3. RECURSE: Find children inside the body
                elif any((kind.startswith(x) for x in ("Instance", "GenerateBlock",
                                                       "Package"))):
                    if module_type and kind == 'Package':
                        new_traversed_modules = [module_type]
                    elif module_type:
                        new_traversed_modules = traversed_modules + [module_type]
                    else:
                        new_traversed_modules = traversed_modules

                    if max_depth > 0:
                        child_results.extend(self.get_items(
                            node=m, path=current_path,
                            traversed_modules=new_traversed_modules,
                            max_depth=(max_depth-1),
                            only_top_module=only_top_module,
                            only_paths=only_paths,
                            only_ports=only_ports, only_module_types=only_module_types,
                            only_covergroups=only_covergroups, only_files=only_files
                        ))


            if collect and node_kind == "Instance":
                if module_type:
                    new_traversed_modules = traversed_modules + [module_type]
                else:
                    new_traversed_modules = traversed_modules
                self_data = [{
                    "path": current_path,
                    "module": module_type,
                    "traversed_modules": new_traversed_modules,
                    "inst_name": inst_name,
                    "file": str(Path(source_file).absolute()),
                    "ports": ports,
                    "signals": signals,
                    "covergroups": covergroups
                }]
            else:
                self_data = []

            return self_data + child_results



        # 4. Handle the top-level design container
        results = []
        # If the current node isn't an instance but has members (like the root), check them
        for m in node.get("members", []):
            results.extend(self.get_items(
                node=m, path=path,
                traversed_modules=traversed_modules,
                max_depth=max_depth,
                only_top_module=only_top_module,
                only_paths=only_paths,
                only_ports=only_ports, only_module_types=only_module_types,
                only_covergroups=only_covergroups, only_files=only_files
            ))

        if results:
            return results

        if m := node.get('design', {}):
            results.extend(self.get_items(
                node=m, path=path,
                traversed_modules=[],
                max_depth=max_depth,
                only_top_module=only_top_module,
                only_paths=only_paths,
                only_ports=only_ports, only_module_types=only_module_types,
                only_covergroups=only_covergroups, only_files=only_files
            ))

        return results



'''
Windows utils for simple things in other OSes, such as: finding anything in
$env:LOCALAPPDATA (in case App exe is not in $env:Path)
'''

import os
from opencos import eda_config
from opencos.utils import subprocess_helpers

IS_WINDOWS = subprocess_helpers.IS_WINDOWS


def get_windows_tool_info( # pylint: disable=dangerous-default-value
        app_name: str, prefer_exes: list = []
) -> dict:
    """
    Resolves the actual binary path and version from a Windows electron app.
    app_name: e.g., 'Cognichip'

    Will look in:
      -- $LOCALAPPDATA
      -- $PROGRAMFILES
      -- ./Programs/ for each
    """
    ret = {}
    if not IS_WINDOWS:
        return ret

    local_app_data = os.environ.get('LOCALAPPDATA', '')
    program_files = os.environ.get('PROGRAMFILES', '')

    try_dirs = [
        local_app_data,
        os.path.join(local_app_data, 'Programs'),
        program_files,
        os.path.join(program_files, 'Programs')
    ]

    base_path = ''
    for x in try_dirs:
        if x:
            base_path = os.path.join(x, app_name)
            if os.path.isdir(base_path):
                break

    ret.update({
        'base_path': {
            'value': base_path,
            'exists': os.path.exists(base_path)
        }
    })

    if not os.path.exists(base_path):
        return ret

    # 1. Resolve Executable (Replaces shutil.which), assumes ./bin/ and ./
    # we prefer something in ./bin/tool.cmd, those tend to swallow a few
    # otherwise stdout/stderr on startup.
    exe_path = ''
    found_exe = False
    for _try_dir in ['bin', '']:
        if found_exe:
            break
        for _try_exe_name in prefer_exes + [app_name, f'{app_name}.EXE', f'{app_name}.CMD']:
            exe_path = os.path.join(base_path, _try_dir, _try_exe_name)
            if os.path.exists(exe_path) and \
               os.access(exe_path, os.F_OK) and \
               os.access(exe_path, os.X_OK):
                found_exe = True
                break

    if not found_exe:
        exe_path = ''

    # 2. Resolve Version via product.json
    product_json_path = os.path.join(base_path, 'resources', 'app', 'product.json')

    # We can get the exe shortname from product.json:
    # key = win32ShellNameShort
    # but that doesn't tell us it's in ./bin

    ret.update({
        'product_json_path': {
            'value': product_json_path,
            'exists': os.path.exists(product_json_path)
        },
    })

    # We can't get the version yet unless we open the product_json_path
    ret.update({
        "bin": exe_path,
        "metadata": product_json_path if os.path.exists(product_json_path) else ''
    })
    return ret


def get_tool_exe_from_config(config: dict, tool: str) -> str:
    """
    Returns exe path for 'tool', looking in LOCALAPPDATA, aka
    C:\\Users\\$USER\\AppData\\Local

    If not windows, reverts to the first exe in config.
    tool -- this is the opencos-eda tool name, such as 'verilator',
    from the --tool=TOOL arg, or auto-found.
    """

    tool_cfg = config.get('tools', {}).get(tool, {})
    exe_list = eda_config.tool_get_exe_list(tool=tool, config=config)
    first_exe = exe_list[0]

    if IS_WINDOWS:
        app_name = tool_cfg.get('windows-app-name', '')
        if not app_name:
            app_name = tool_cfg.get('app-name', '')
        if not app_name:
            app_name = tool_cfg.get('macos-app-name', '')

        prefer_exes = tool_cfg.get('windows-prefer-exes', [])
        if exe_list:
            prefer_exes = exe_list + prefer_exes
        if app_name:
            data = get_windows_tool_info(app_name, prefer_exes=prefer_exes)
            if e := data.get('bin', ''):
                first_exe = e

    return first_exe



'''opencos.commands.multi - command handler for: eda multi ..., and eda tools-multi ...

These are not intended to be overriden by child classes. They do not inherit Tool classes.
'''

import argparse
import glob
import os
from pathlib import Path

from rich.pretty import pprint

from opencos import util, eda_base, eda_config, export_helper, \
    eda_tool_helper
from opencos.deps.deps_file import get_deps_markup_file, deps_markup_safe_load, \
    deps_data_get_all_targets, deps_list_target_sanitize
from opencos.eda_base import CommandParallel, get_eda_exec
from opencos.files import safe_shutil_which
from opencos.util import Colors
from opencos.utils.str_helpers import fnmatch_or_re, dep_str2list


class CommandMulti(CommandParallel):
    '''eda.py command handler for: eda multi <command> <args,targets,target-globs,...>'''

    command_name = 'multi'
    subcommand = None # Note that self.single_command / command is NOT assigned to subcommand

    def __init__(self, config: dict):
        CommandParallel.__init__(self, config=config)

        self.multi_only_args = {
            'fake': False,
            'single-timeout': None,
            'slurm-srun': False,
            'parallel': 1,
            'threads': util.get_thread_budget(),
            'threads-per-child': 1,
            'use-max-threads': False,
            'fail-if-no-targets': True,
            'export-jsonl': False,
            'print-targets': False,
            # TODO(drew): coverage args disabled for now
            # These should probably only be added for `eda multi sim`
            #'coverage-merge': False,
            #'coverage-merge-output': 'eda.work/merged.acdb',
            #'coverage-merge-report': True,
        }

        self.args.update(self.multi_only_args)
        self.args_help.update({
            'single-timeout': ('shell timeout on a single operation in multi, not the entire'
                               'multi command'),
            'slurm-srun': (
                'Run each single job through slurm via: srun eda COMMAND ...;'
                ' consider using with --parallel=P (P > 1). Note this does not yet'
                ' work with eda --ssh=name@host'
            ),
            'fail-if-no-targets': 'fails the multi command if no targets were found',
            'export-jsonl': ('If set, generates export.jsonl if possible, spawns single commands'
                             'with --export-json'),
            'print-targets': 'Do not run jobs, prints targets to stdout',
            # TODO(drew): coverage args disabled for now
            #'coverage-merge': (
            #    'Enable automatic coverage database merging after parallel test runs. '
            #    'Requires --coverage flag in the command arguments.'
            #),
            #'coverage-merge-output': (
            #    'Output path for merged coverage database (default: eda.work/merged.acdb)'
            #),
            #'coverage-merge-report': (
            #    'Auto-generate coverage reports after merging (default: True). '
            #    'Use --no-coverage-merge-report to disable.'
            #),
        })

        self.args_args['threads'] = ['-j'] # support -j | --threads

        self.single_command = ''
        self.targets = [] # list of tuples (target:str, tool:str)
        self.resolve_target_command = ''
        self.arg_tokens = [] # stealthily used by _detect_coverage_enabled(self)


    def path_hidden_or_work_dir(self, path: str) -> bool:
        '''Returns True if any portion of path is hidden file/dir or has a work-dir

        such as  "eda.work" (self.args['eda-dir'])'''

        path_obj = Path(os.path.abspath(path))
        if self.args['eda-dir'] in path_obj.parts:
            return True
        if any(x.startswith('.') and len(x) > 1 for x in path_obj.parts):
            return True
        return False

    def _detect_coverage_enabled(self) -> bool:
        '''Check if --coverage flag is present in arg_tokens passed to child commands'''
        for token in getattr(self, 'arg_tokens', []):
            if token.startswith('--coverage'):
                return True
        return False

    def collect_coverage_databases(self) -> list:
        '''Collect work.acdb files from completed jobs.
        Pattern: {work-dir}/{short_target}.{command}/work.acdb (single tool)
                 {work-dir}/{short_target}.{command}.{tool}/work.acdb (multi tool)
        Returns list of absolute paths to valid .acdb files.
        '''
        coverage_files = []
        command = self.single_command

        work_dir = self.args.get('work-dir') or self.args.get('eda-dir', 'eda.work')

        for target, tool in self.targets:
            _, short_target = os.path.split(target)
            # Try both patterns since the actual pattern depends on job naming
            #     pattern 1: eda.work/{target}.{command}/work.acdb
            #     pattern 2: eda.work/{target}.{command}.{tool}/work.acdb
            patterns = [
                os.path.join(work_dir, f'{short_target}.{command}', 'work.acdb'),
                os.path.join(work_dir, f'{short_target}.{command}.{tool}', 'work.acdb'),
            ]

            for work_pattern in patterns:
                if os.path.exists(work_pattern) and work_pattern not in coverage_files:
                    coverage_files.append(os.path.abspath(work_pattern))

        return coverage_files

    # TODO(drew): coverage disabled for now, this should be refactored to
    # be called via `eda coverage merge` instead of creating a handler class
    #def merge_coverage_databases(self, coverage_files: list):
    #    '''Invoke coverage merge using the tool-specific handler from config.
    #    '''
    #    coverage_handler_cls = self.config.get('command_handler', {}).get('coverage')
    #    if not coverage_handler_cls:
    #        util.warning("No coverage handler configured. Ensure the active tool supports "
    #                     "coverage merging (check --tool and config).")
    #        return
    #
    #    merge_cmd = coverage_handler_cls(config=self.config)
    #    merge_cmd.coverage_files = coverage_files
    #    merge_cmd.subcommand = 'merge'
    #    merge_cmd.args['output'] = self.args['coverage-merge-output']
    #    merge_cmd.args['coverage-report'] = self.args['coverage-merge-report']
    #    for arg_name, _ in COVERAGE_BASELINE_ARGS.items():
    #        merge_cmd.args[arg_name] = self.args.get(
    #            arg_name, COVERAGE_BASELINE_ARGS[arg_name]
    #        )
    #
    #    try:
    #        merge_cmd.do_it()
    #    except Exception as e:
    #        util.warning(f"Coverage merge failed: {e}")


    def resolve_target_get_command_level(self, level: int = -1) -> (str, int):
        '''increments level, returns tuple of (command str, level int)'''
        command = self.resolve_target_command
        if level < 0:
            level = 1
        else:
            level += 1
        return command, level


    def resolve_path_and_target_patterns( # pylint: disable=too-many-locals
            self, base_path: str, target: str, level: int = -1
    ) -> dict:
        '''Returns a dict of: key = matching path, value = set of matched targets.

        Looks at globbed paths from base_path/target, and looks for DEPS markup targets
        matching target_pattern (using fnmatch or re.match)
        '''
        def debug(*text):
            util.debug(f'resolve_target() {level=} {base_path=}', *text)

        # join base_path / target
        #   - if target = ./some_path/**/*test
        # split them again so we can do all globbing on the path_pattern portion:
        path_pattern, target_pattern = os.path.split(os.path.join(base_path, target))
        debug(f'{path_pattern=}, {target_pattern=}')

        if target_pattern == '...':
            # replace bazel style target ... with '*' for fnmatch.
            target_pattern = '*'

        matching_targets_dict = {}

        # Let's not glob.glob if the path_pattern and target_pattern are
        # exact, aka if it does not have special characters for glob: *, ?, []
        # for the target, we also support re, so: * + ?
        if any(x in path_pattern for x in ('*', '?', '[')):
            paths_from_pattern = list(glob.glob(path_pattern, recursive=True))
        else:
            paths_from_pattern = [path_pattern]

        target_pattern_needs_lookup = any(x in target_pattern for x in ['*', '?', '+'])

        # resolve the path_pattern portion using glob.
        # we'll have to check for DEPS markup files in path_pattern, to match the target_wildcard
        # using fnmatch or re.
        for path in paths_from_pattern:

            if self.path_hidden_or_work_dir(path):
                continue

            deps_markup_file = get_deps_markup_file(path)
            if deps_markup_file:
                data = deps_markup_safe_load(deps_markup_file)
                deps_targets = deps_data_get_all_targets(data)
                rel_path = os.path.relpath(path)

                debug(f'in {rel_path=} looking for {target_pattern=} in {deps_targets=}')

                for t in deps_targets:
                    if target_pattern_needs_lookup:
                        matched = fnmatch_or_re(pattern=target_pattern, string=t)
                    else:
                        matched = t == target_pattern
                    if matched:
                        if rel_path not in matching_targets_dict:
                            matching_targets_dict[rel_path] = set()
                        matching_targets_dict[rel_path].add(t)

        debug(f'Found potential targets for {target_pattern=}: {matching_targets_dict=}')
        return matching_targets_dict


    def resolve_target(self, base_path: str, target: str, level: int = -1) -> None:
        '''Returns None, recursively attempts to determine the validity of a base_path/target,

        and appends to self.targets. self.resolve_target_command (str) and level (int) exist
        for debug messaging. Auto-increments 'level' arg, so caller can invoke with
        level=(their_current_level).
        '''

        def debug(*text):
            util.debug(f'resolve_target() {level=} {base_path=}', *text)

        command, level = self.resolve_target_get_command_level(level)

        debug(f"Enter/Start: target={target}, command={command}")

        # Strip outer quote on target, in case it was passed this way from CLI:
        for x in ['"', "'"]:
            target = target.lstrip(x).rstrip(x)

        matching_targets_dict = self.resolve_path_and_target_patterns(
            base_path=base_path, target=target, level=level
        )

        for path, targets in matching_targets_dict.items():
            self.resolve_target_single_path(base_path=path, targets=targets, level=level)


    def resolve_target_single_path( # pylint: disable=too-many-locals
            self, base_path: str, targets: list, level: int = -1
    ) -> None:
        '''Returns None, called by resolve_target(..) if we have a single base_path,

        and multiple targets (list), and need to resolve it via base_path and DEPS.[markup]
        file information. There should be no remaining wildcard information in targets
        (that was handled earlier with fnmatch and re.)
        '''

        def debug(*text):
            util.debug(f'resolve_target() {level=} {base_path=}', *text)

        command, level = self.resolve_target_get_command_level(level)

        all_multi_tools = self.multi_which_tools(command)

        deps_file = get_deps_markup_file(base_path)
        data = {}
        if self.config['deps_markup_supported'] and deps_file:
            data = deps_markup_safe_load(deps_file)

        deps_targets = deps_data_get_all_targets(data)
        deps_file_defaults = data.get('DEFAULTS', {})

        # Loop through all the targets in DEPS.yml, skipping DEFAULTS
        for target_node in targets:
            if target_node not in deps_targets:
                continue

            entry = data[target_node]

            # Since we support a few schema flavors for a target (our
            # 'target_node' key in a DEPS.yml file) santize the entry
            # so it's a {} with a 'deps' key:
            entry_sanitized = deps_list_target_sanitize(
                entry, target_node=target_node, deps_file=deps_file
            )

            # Start with the defaults, and override with this entry_sanitized
            entry_with_defaults = deps_file_defaults.copy()
            entry_with_defaults.update(entry_sanitized)
            entry = entry_with_defaults

            # Because CommandMulti has child CommandToolsMulti, we support multiple tools
            # and have multi ignore waivers tha may only be some of those tools. Keep a list
            # of tools we skip for this target.
            multi_ignore_skip_this_target_node = set() # which tools we'll skip

            # Check if this target_node should be skipped due to:
            # multi - ignore-this-target (commands or tools)
            multi_ignore_commands_list = entry.get('multi', {}).get(
                'ignore-this-target', []
            )

            for x in multi_ignore_commands_list:
                if len(multi_ignore_skip_this_target_node) == len(all_multi_tools):
                    # If we already found a reason to not use this target due to multi - ignore,
                    # on all tools, then stop.
                    break

                assert isinstance(x, dict), \
                    (f'multi ignore-this-target: {x=} {multi_ignore_commands_list=}'
                     f' {deps_file_defaults=} This needs to be a dict entry with keys'
                     f'"commands" and "tools" {deps_file=} {target_node=}')

                commands = x.get('commands', [])
                tools = x.get('tools', [])
                ignore_commands_list = dep_str2list(commands)
                ignore_tools_list = dep_str2list(tools)

                debug(f"{ignore_tools_list=}, {ignore_commands_list=} {target_node=}")
                debug(f"{command=} --> {all_multi_tools=}")
                if not ignore_commands_list or \
                   command in ignore_commands_list or \
                   ignore_commands_list == ['None']:
                    # if commands: None, or commands is blank, then assume it is all commands.
                    # (note that yaml doesn't support *)

                    for tool in all_multi_tools:
                        if tool in ignore_tools_list or ignore_tools_list == ['None'] or \
                           len(ignore_tools_list) == 0:
                            # if tools: None, or is blank, then assume it is for all tools
                            debug(f"Skipping {target_node=} due to using {command=} {tool=}",
                                  f"given {ignore_tools_list=} and {ignore_commands_list=}")
                            multi_ignore_skip_this_target_node.add(tool)

            for tool in all_multi_tools:
                if tool not in multi_ignore_skip_this_target_node:
                    debug(f"Found dep {target_node=} {tool=} matching, {entry=}")
                    self.targets.append( tuple([os.path.join(base_path, target_node), tool]) )


    def process_tokens( # pylint: disable=too-many-locals, too-many-branches, too-many-statements
            self, tokens: list, process_all: bool = True, pwd: str = os.getcwd()
    ) -> list:
        '''CommandMulti.process_tokens(..) is likely the entry point for: eda multi <command> ...

        - handles remaining CLI arguments (tokens list)
        - builds list of jobs in self.jobs and runs them.
        '''
        # multi is special in the way it handles tokens, due to most of them being processed by
        # a subprocess to another eda.Command class (for the command)
        arg_tokens = [] # these are the tokens we will pass to the child eda processes
        command = ""
        target_globs = []
        tool = None
        orig_tokens = tokens.copy()

        # We want to run our built-in argparser on self.args keys, but that would end up
        # parsing args we'd like to pass to individual eda commands from multi, so instead
        # only run it on a subset of our self.args:
        _, unparsed = self.run_argparser_on_list(
            tokens=tokens,
            parser_arg_list=list(self.multi_only_args.keys()),
            apply_parsed_args=True
        )

        command = self.get_sub_command_from_config()

        # Need to know the tool for this command, either it was set correctly via --tool and/or
        # the command (class) will tell us.
        all_multi_tools = self.multi_which_tools(command)

        single_cmd_unparsed = self.get_unparsed_args_on_single_command(
            command=command, tokens=unparsed
        )

        util.debug(f"Multi: {unparsed=}, looking for target_globs")
        for token in unparsed:
            if token in single_cmd_unparsed:
                target_globs.append(token)
            else:
                arg_tokens.append(token)

        self.arg_tokens = arg_tokens

        # now we need to expand the target list
        self.single_command = command
        util.debug(f"Multi: {orig_tokens=}")
        util.debug(f"Multi: {command=}")
        util.debug(f"Multi: {self.config=}")
        util.debug(f"Multi: {all_multi_tools=}")
        util.debug(f"Multi: {target_globs=}")
        util.debug(f"Multi: {arg_tokens=}")
        if self.args.get('export-jsonl', False):
            util.info("Multi: --export-jsonl")
        self.targets = []
        cwd = util.getcwd()
        current_targets = 0
        self.resolve_target_command = command
        for target in target_globs:
            self.resolve_target(base_path=cwd, target=target)
            if len(self.targets) == current_targets:
                # we didn't get any new targets, try globbing this one
                for f in glob.glob(target):
                    if os.path.isfile(f):
                        util.info(f"Adding raw file target: {f} for tools {all_multi_tools}")
                        for tool in all_multi_tools:
                            self.targets.append( tuple([f, tool]) )
            current_targets = len(self.targets)
        util.info(f"Multi: Expanded {target_globs} to {len(self.targets)} {command} targets")

        if self.args['fail-if-no-targets'] and not self.targets:
            self.error(f'Multi: --fail-if-no-targets set, and {self.targets=}. Disable with',
                       '--no-fail-if-no-targets, or see: eda multi --help')
        if not all_multi_tools:
            possible_tools = self.all_handler_commands.get(command, [])
            self.error(f'Multi: no tools to run for {command=}, available tools: {possible_tools}')

        if user_host := self.config.get('ssh'):
            util.info(f'{Colors.off_bgreen}Running via ssh -- {Colors.bcyan}{user_host}')

        util.info("Multi: About to run: ", end="")

        def get_pretty_targets_tuple_as_list(l:list):
            # prints 'a(b)', used for 'target(tool)' for tuples in self.targets
            if len(all_multi_tools) > 1:
                return [f'{a.split("/")[-1]}({b})' for a,b in l]

            # don't add the (tool) part if we're only running 1 tool
            return [f'{a.split("/")[-1]}' for a,b in l]

        if len(self.targets) > 20:
            mylist = get_pretty_targets_tuple_as_list(self.targets[:10])
            util.info( ", ".join(mylist), start="", end="")
            util.info( ", ... ", start="", end="")
            mylist = get_pretty_targets_tuple_as_list(self.targets[-10:])
            util.info( ", ".join(mylist), start="")
        else:
            mylist = get_pretty_targets_tuple_as_list(self.targets)
            util.info( ", ".join(mylist), start="")

        if self.args['print-targets']:
            util.info('Multi print-targets (will not run jobs): -->')
            print_targets = [t[0] for t in self.targets]
            print_targets.sort()
            for x in print_targets:
                # t = tuple of (target:str, tool:str), we just want the target.
                print(f'  {x}')
        else:
            util.debug("Multi: converting list of targets into list of jobs")
            self.jobs = []
            self.append_jobs_from_targets(args=arg_tokens)
            self.run_jobs(command)


            # TODO(drew): coverage disabled for now
            # safe wrapper to call self.merge_coverage_databases:
            #if self.args.get('coverage-merge', False) and self._detect_coverage_enabled():
            #    util.debug(
            #        f"Coverage merge: work-dir={self.args['work-dir']},",
            #        f"targets={len(self.targets)}"
            #    )
            #    coverage_files = self.collect_coverage_databases()
            #    util.debug(
            #        f"Coverage merge: collected {len(coverage_files)} files: {coverage_files}"
            #    )
            #    if coverage_files:
            #        util.info(f"Merging {len(coverage_files)} coverage databases...")
            #        self.merge_coverage_databases(coverage_files)
            #    else:
            #        util.warning("--coverage-merge enabled but no coverage databases found")

        # Because CommandMulti has a custom arg parsing, we do not have 'export' related
        # args in self.args (they are left as 'unparsed' for the glob'ed commands)
        # Note that --export-jsonl has already been removed from 'unparsed' and is in parsed,
        # and would already be set in self.args.
        bool_action_kwargs = util.get_argparse_bool_action_kwargs()
        export_parser = argparse.ArgumentParser(prog='eda', add_help=False, allow_abbrev=False)
        for arg,v in self.args.items():
            if arg.startswith('export') and isinstance(v, bool):
                export_parser.add_argument(f'--{arg}', **bool_action_kwargs)
        try:
            export_parsed, export_unparsed = export_parser.parse_known_args(unparsed + [''])
            unparsed = list(filter(None, export_unparsed))
        except argparse.ArgumentError:
            self.error(f'problem attempting to parse_known_args for {unparsed=}')

        for key,value in vars(export_parsed).items():
            if key not in self.args and '_' in key:
                # try with dashes instead of _
                key = key.replace('_', '-')
            if value is None:
                continue
            self.args[key] = value # set one of the parsed 'export' args
            util.info(f'Export: setting arg {key}={value}')

        if self.is_export_enabled():
            self.do_export()


    def which_tool(self, command):
        # Do not use for CommandMulti, b/c we support list of tools.
        raise NotImplementedError


    def multi_which_tools(self, command) -> list:
        '''returns a list, or None, of the tool that was already determined to run the command

        CommandToolsMulti will override and return its own list'''
        if self.config.get('ssh') and self.config.get('ssh_tool'):
            return [self.config['ssh_tool']]
        return [eda_base.which_tool(command, config=self.config)]


    def _append_job_command_args( # pylint: disable=R0913,R0917 # too-many-arguments
            self, command_list: list, tool: str, all_multi_tools: list, short_target: str,
            command: str
    ) -> None:
        '''Returns a dict of everything we added'''

        added = super().update_args_list(args=command_list, tool=tool)
        if self.args.get('export-jsonl', False):
            # Special case for 'multi' --export-jsonl, run reach child with --export-json
            added['export-json'] = ''
            command_list.append('--export-json')
        if tool and len(all_multi_tools) > 1:
            jobname = f'{short_target}.{command}.{tool}'
        else:
            jobname = f'{short_target}.{command}'
        added['job-name'] = jobname
        command_list.append(f'--job-name={jobname}')
        logfile = os.path.join(self.args['eda-dir'], f'eda.{jobname}.log')
        added['logfile'] = logfile
        command_list.append(f'--force-logfile={logfile}')
        return added


    def append_jobs_from_targets(self, args:list):
        '''Helper method in CommandMulti to apply 'args' (list) to all self.targets,

        and add to self.jobs (list) to be run later.
        '''
        eda_path = get_eda_exec('multi')
        command = self.single_command
        timeout = safe_shutil_which('timeout')
        srun = safe_shutil_which('srun')

        # Built-in support for running > 1 tool.
        all_multi_tools = self.multi_which_tools(command)
        util.info(f'Multi - append_jobs_from_targets: {command=} {all_multi_tools=}')

        for target, tool in self.targets:
            command_list = [
                eda_path, command, '--threads=' + str(self.get_eda_child_thread_budget())
            ]

            assert target, f'{target=} {tool=}'

            _, short_target = os.path.split(target) # trim path info on left

            added_args_info = self._append_job_command_args(
                command_list=command_list, tool=tool, all_multi_tools=all_multi_tools,
                short_target=short_target, command=command
            )

            # If invoked with --ssh, forward it to each child so the target runs remotely.
            # Each child gets its own ephemeral remote sandbox (ExportHelper + tar/SSH).
            if self.config.get('ssh'):
                command_list += [f'--ssh={self.config["ssh"]}']
                if self.config.get('ssh_setup'):
                    command_list += ['--ssh-setup', self.config['ssh_setup']]

            # if self.args['parallel']: command_list += ['--quiet']
            command_list += args # put the args prior to the target.
            command_list += [target]

            # prepend a nix-style 'timeout <seconds>' on the command_list if this was set
            # NOTE: it might make more sense to merge --single-timeout into the
            # python --subprocess-timeout, although this one survives to the generated
            # .sh scripts in the work-dir.
            if timeout and \
               self.args.get('single-timeout', None) and \
               type(self.args['single-timeout']) in [int, str]:
                command_list = ['timeout', str(self.args['single-timeout'])] + command_list

            # TODO(drew): if you wanted to run --ssh + --slurm-srun, we'd have to send the
            # single job via ssh, and then the single job (eda_base.Command) would have to
            # support --slurm-srun to prepend "srun".

            # handle --slurm-srun arg, run `srun` in from of the command:
            if srun and self.args.get('slurm-srun', False):
                command_list = ['srun'] + command_list

            name = self.get_name_from_target(target)
            if tool and (len(all_multi_tools) > 1 or self.command_name == 'tools-multi'):
                name += f' ({tool})'

            this_job_dict = {
                'job-name': added_args_info['job-name'],
                'name' : name,
                'index' : len(self.jobs),
                'command': command,
                'target': target,
                'command_list' : command_list
            }
            if tool:
                util.debug(f'Multi: append_jobs_from_targets: {tool=} this_job_dict:')
                if util.args['debug']:
                    pprint(this_job_dict)
            else:
                util.debug('Multi: append_jobs_from_targets: this_job_dict:')
                if util.args['debug']:
                    pprint(this_job_dict)
            self.jobs.append(this_job_dict)


    def do_export(self, **kwargs) -> None:
        '''Override from Command.do_export -- Perform export for a multi/tools-multi command

        For an example command:
            eda multi <command> --export[-jsonl] <args,targets,...>
        '''
        if self.args.get('work-dir', '') == '':
            self.args['work-dir'] = self.args.get('eda-dir', 'eda.work')

        util.info('Multi export: One of the --export[..] flag set, may examine',
                  f'{self.args["work-dir"]=}')
        self.collect_single_exported_export_jsonl()
        util.info('Mulit export: done')


    def collect_single_exported_export_jsonl(self) -> None:
        '''Create a single JSONL or JSON file for all multi jobs'''
        do_as_jsonl = self.args.get('export-jsonl', False)
        do_as_json = self.args.get('export-json', False)

        if not do_as_json and not do_as_jsonl:
            return

        if do_as_jsonl:
            outfile_str = 'export.jsonl'
        else:
            outfile_str = 'export.json'

        command = self.single_command
        all_multi_tools = self.multi_which_tools(command)

        json_file_paths = []
        for target, tool in self.targets:
            # Rather than glob out ALL the possible exported files in our work-dir,
            # only look at the multi targets:
            p, target_nopath = os.path.split(target)
            if not target_nopath:
                target_nopath = p # in case self.targets was missing path info

            if len(all_multi_tools) > 1:
                # Need to look in:
                #   eda.work/<shorttarget>.<command>.<tool>/export/export.json
                # If this was 'eda export' command, then need to look in:
                #   eda.export/......./export.json.
                single_pathname = os.path.join(
                    self.args['work-dir'],
                    f'{target_nopath}.{self.single_command}.{tool}',
                    'export', 'export.json'
                )
            else:
                # We only ran for 1 tool, so the tool value is a dontcare in the output path
                # Need to look in eda.work/<shorttarget>.<command>/export/export.json
                single_pathname = os.path.join(
                    self.args['work-dir'],
                    f'{target_nopath}.{self.single_command}',
                    'export', 'export.json'
                )
            util.debug(f'Looking for export.json in: {single_pathname=}')
            if os.path.exists(single_pathname):
                json_file_paths.append(single_pathname)


        output_json_path = os.path.join(self.args['work-dir'], 'export', outfile_str)
        if len(json_file_paths) == 0:
            self.error(f'{json_file_paths=} is empty list, no targets found to export',
                       f'for {output_json_path=}')
            return

        # TODO(drew): If we ran this w/ several tools from CommandToolsMulti, we'll end up
        # with tests having same name (but different tool). Might need to uniquify the names.
        util.debug(f'Multi export: {json_file_paths=}')
        if do_as_jsonl:
            util.info(f'Multi export: saving JSONL format to: {output_json_path=}')
            export_helper.json_paths_to_jsonl(json_file_paths=json_file_paths,
                                              output_json_path=output_json_path)
        else:
            util.info('Multi export: saving JSON format to: {output_json_path=}')
            export_helper.json_paths_to_single_json(json_file_paths=json_file_paths,
                                                    output_json_path=output_json_path)




class CommandToolsMulti(CommandMulti):
    '''eda.py command handler for: eda tools-multi <args,targets,target-globs,...>

    This class is used to support running a multi-style command, but on many
    tools, such as:
       eda tools-multi sim --tools=verilatr --tools=iverilog *test
    '''

    command_name = 'tools-multi'

    def __init__(self, config: dict):
        super().__init__(config=config)
        self.all_handler_commands = {} # cmd: [ordered list of tools]
        self.tools = set()
        self.args.update({
            'tools': [], # Used for help, will internally use self.tools from argparser.
        })
        self.args_help.update({
            'tools': 'list of tools to run for eda multi targets, such as' \
            + ' --tools=modelsim_ase --tools=verilator=/path/to/bin/verilator',
        })
        if 'tool' in self.args:
            self.args.pop('tool')

        self.update_all_known_tools()

    def update_all_known_tools(self):
        '''Checks whats tools are loaded by opencos.eda, and updates the

        command handlers. This is necessary otherwise we may not run the correct
        tools in a (eda tools-multi <command> --tools=tool1 --tools=tool2 ...) style
        command
        '''
        cfg, tools_loaded = eda_tool_helper.get_config_and_tools_loaded(quiet=True)
        self.all_handler_commands = eda_tool_helper.get_all_handler_commands(cfg, tools_loaded)
        util.debug('CommandToolsMulti: all_handler_commands:')
        if util.args['debug']:
            pprint(self.all_handler_commands)

    def multi_which_tools(self, command):
        '''Overrides CommandMulti.multi_which_tool(command), return a list of all
        possible tools that can run this command'''
        if self.tools is None or not self.tools:
            # wasn't set via arg --tools, so use all if possible for this command.
            which_tools = self.all_handler_commands.get(command, [])
        else:
            # self.tools set from args --tools (list)
            which_tools = [tool for tool in self.all_handler_commands.get(command, []) \
                           if tool in self.tools]
        return which_tools

    def process_tokens(self, tokens: list, process_all: bool = True,
                       pwd: str = os.getcwd()) -> list:

        # setup an argparser to append tools to a list, if no tools set, then use
        # all possible tools (only do this for arg '--tools'):
        parsed, unparsed = self.run_argparser_on_list(
            tokens=tokens,
            parser_arg_list=[
                'tools',
            ],
            apply_parsed_args=False
        )

        if not parsed.tools:
            self.tools = set()
        else:
            # deal with --tools=name=/path/to/name (update config w/ path info):
            self.tools = set(
                eda_config.update_config_auto_tool_order_for_tools(
                    tools=parsed.tools, config=self.config
                )
            )
            util.info(f'CommandToolsMulti: {self.tools=}')
        self.args['tools'] = self.tools

        # Call ComamndMulti's process_tokens:
        return super().process_tokens(
            tokens=unparsed, process_all=process_all, pwd=pwd
        )

    def which_tool(self, command):
        # Do not use for CommandMulti, b/c we support list of tools.
        raise NotImplementedError



'''Various checks for making sure a user can or can't run docker

Uses:
from opencos.utils.docker_checks import docker_ok

if docker_ok():
  # do docker stuff
else
  # print messages saying can't do docker stuff, fail
'''

from importlib import import_module
import os
from pathlib import Path
import shutil
import subprocess
import sys

from opencos import __version__ as opencos_version
from opencos.util import Colors, debug, warning, warn_once, info_once, get_thread_budget
from opencos.utils.str_helpers import strip_outer_quotes
from opencos.utils.subprocess_helpers import IS_LINUX, IS_MACOS, IS_WINDOWS

DOCKER_EXE = shutil.which('docker')
DOCKER_GROUP = 'docker'

_DOCKER_VERSION = ''
_DOCKER_OK = None
_CAN_RUN_DOCKER_WITHOUT_SUDO = None


grp = None # pylint: disable=invalid-name
if IS_LINUX:
    grp = import_module('grp') # pylint: disable=invalid-name


def get_error_str_cant_run_docker() -> str:
    '''Returns a somewhat O/S specific messsage about you couldn't run docker'''

    is_installed = bool(DOCKER_EXE)
    if not is_installed:
        return (
            'Current user cannot run docker, it does not appear to be installed or is not in PATH'
        )
    if not not_in_docker_already():
        return (
            'Current user cannot run docker, it appears we are already in a docker ENV'
            ' (./dockerenv exists). Docker-in-docker is not yet supported within opencos-eda.'
        )
    if IS_LINUX:
        return (
            'Current user cannot run docker! (user is not a member of',
            ' docker group, or requires sudo to run docker).'
        )
    if IS_MACOS:
        return (
            'Current macos user cannot run docker! (user may not have Docker Desktop'
            ' running, it needs to be. Also check if user can run docker without sudo.)'
        )
    if IS_WINDOWS:
        return (
            'Current windows user cannot run docker! (This may not be correctly implemented'
            f' for windows in opencos-eda version: {opencos_version})'
        )
    return (
        'Current user cannot run docker! (unknown operating system, not supported in'
        f' opencos-eda version: {opencos_version} -- platform: {sys.platform})'
    )

def docker_info_for_pull(
        tool: str, image: str, name: str = '_docker_pull_image', **kwargs
) -> None:
    '''
    A standard info for tools that may optionally (not required to) support docker,
    just as a reminder we don't have version information b/c image not pulled
    '''
    txt = (
        f'For {Colors.off}{Colors.cyan}--tool={tool}{Colors.info}, the default image is not',
        f'present on this system. Use: {Colors.off}{Colors.yellow}docker pull {image}{Colors.info}',
        f'or run with a different image using {Colors.off}{Colors.cyan}--docker-image=',
        f'{Colors.yellow}IMAGE{Colors.info},',
        f'see {Colors.off}{Colors.cyan}--help{Colors.info} for other options.'
    )
    info_once(*txt, info_name=f'{tool}{name}', **kwargs)


def docker_info_for_tool(
        tool: str, image: str, name: str = '_get_versions_docker', is_warning: bool = False,
        **kwargs
) -> None:
    '''
    A standard info for tools that may optionally (not required to) support docker,
    just as a reminder that they can run with --docker-run and --docker-image
    '''
    txt = (
        f'{Colors.off}{Colors.cyan}"{tool}"{Colors.info} not in path. To run with docker, use:',
        f'{Colors.off}{Colors.cyan}--docker-run{Colors.info}',
        f'{Colors.off}{Colors.cyan}--docker-image={Colors.yellow}{image}{Colors.info},',
        f'see {Colors.off}{Colors.cyan}--help{Colors.info} for other options.'
    )
    if is_warning:
        warn_once(*txt, warn_name=f'{tool}{name}', **kwargs)
    else:
        info_once(*txt, info_name=f'{tool}{name}', **kwargs)


def docker_warning_for_tool(
        tool: str, image: str, name: str = '_get_versions_docker', **kwargs
) -> None:
    '''
    A standard warning for tools that may optionally (not required to) support docker,
    just as a reminder that they can run with --docker-run and --docker-image
    '''
    docker_info_for_tool(tool, image, name, warning=True, **kwargs)



def is_docker_installed() -> bool:
    '''Checks if DOCKER_EXE is non-blank'''
    return bool(DOCKER_EXE)

def not_in_docker_already() -> bool:
    '''Returns True if we're not in docker (False if we are in docker).

    We do not yet support DinD situations, so if we are already in docker,
    make sure docker_ok() returns overall False.
    '''
    return not Path('/.dockerenv').exists()


def get_docker_version() -> str:
    '''Returns docker version (or blank str if not installed) and sets global _DOCKER_VERSION'''
    global _DOCKER_VERSION # pylint: disable=global-statement
    if _DOCKER_VERSION:
        return _DOCKER_VERSION
    if not is_docker_installed():
        return ''

    version_ret = subprocess.run(
        [DOCKER_EXE, '--version'],
        capture_output=True,
        check=False
    )
    # Docker version 29.2.0, build 0b9d198
    stdout = version_ret.stdout.decode('utf-8', errors='replace')
    debug(f'{DOCKER_EXE=} {version_ret=}')
    words = stdout.split()
    if len(words) < 1:
        warning(
            f'{DOCKER_EXE} --version: returned unexpected string {version_ret=}'
        )
    _DOCKER_VERSION = words[2].rstrip(',')
    return _DOCKER_VERSION


def can_run_docker_without_sudo() -> bool:
    '''Checks if the user can run 'docker' with sudo

    Note this command is slow until it has been run and cached.
    '''
    global _CAN_RUN_DOCKER_WITHOUT_SUDO # pylint: disable=global-statement
    if _CAN_RUN_DOCKER_WITHOUT_SUDO is not None:
        return _CAN_RUN_DOCKER_WITHOUT_SUDO
    if not is_docker_installed():
        _CAN_RUN_DOCKER_WITHOUT_SUDO = False

    else:
        try:
            # 'docker info' requires connection to the daemon
            # Using check_call ensures we catch non-zero exit codes
            subprocess.check_call([DOCKER_EXE, 'info'],
                                 stdout=subprocess.DEVNULL,
                                 stderr=subprocess.DEVNULL)
            _CAN_RUN_DOCKER_WITHOUT_SUDO = True
        except (subprocess.CalledProcessError, FileNotFoundError):
            _CAN_RUN_DOCKER_WITHOUT_SUDO = False

    return _CAN_RUN_DOCKER_WITHOUT_SUDO


def is_in_docker_group_or_non_linux() -> bool:
    '''Checks if the user is in a group named 'docker'.

    Note this applies to Linux-only. Macos supports groups but they don't need
    docker group for Docker Desktop'''
    if not IS_LINUX:
        return True
    if not grp:
        return False
    try:
        docker_gid = grp.getgrnam(DOCKER_GROUP).gr_gid
        return docker_gid in os.getgroups() # pylint: disable=no-member
    except KeyError:
        # Group 'docker' doesn't even exist on this system
        return False


def is_docker_ok_windows() -> bool:
    '''Checks for docker support in Windows'''
    if not IS_WINDOWS:
        return True

    return False


def docker_ok() -> bool:
    '''Sets DOCKER_OK to True/False if unset, returns _DOCKER_OK'''
    global _DOCKER_OK # pylint: disable=global-statement
    if _DOCKER_OK is None:
        _DOCKER_OK = all(
            (
                is_docker_installed(),
                not_in_docker_already(),
                get_docker_version(),
                can_run_docker_without_sudo(),
                is_in_docker_group_or_non_linux(),
                is_docker_ok_windows(),
            )
        )
    return _DOCKER_OK


def get_docker_env_from_config(config: dict, tool: str, image: str) -> dict:
    '''Gets the env from config.tools.TOOL for this docker image'''
    tool_cfg = config.get('tools', {}).get(tool, {})
    if not tool_cfg or not image:
        return {}
    if not tool_cfg.get('docker-support', False):
        return {}
    image_no_tag = image.split(':')[0]
    return tool_cfg.get('docker-env', {}).get(image_no_tag, {})


def get_docker_env(cmd_des_obj: object) -> dict:
    '''Gets the env from cmd_des_obj.config.tools.TOOL and cmd_des_obj.args[docker-image]'''
    config = getattr(cmd_des_obj, 'config', None)
    tool = getattr(cmd_des_obj, '_TOOL', None)
    image = getattr(cmd_des_obj, 'args', {}).get('docker-image', '')
    if not config or not tool:
        return {}
    if not image:
        # try the Tool.DEFAULT_DOCKER_IMAGE:
        image = getattr(cmd_des_obj, 'DEFAULT_DOCKER_IMAGE', '')
    if not image:
        return {}
    return get_docker_env_from_config(config=config, tool=tool, image=image)


def get_docker_run_command( # pylint: disable=dangerous-default-value
        image: str,
        docker_command: str = 'run', # or 'exec'
        user: str = '$(id -u):$(id -g)',
        env_dict: dict = {},
        ro_volumes: list = [],
        rw_volumes: list = [],
        entrypoint: str = '/bin/bash',
        commands: list = [],
        threads: int = 1
) -> list:
    '''Returns list (suitable for subprocess) of: docker run ... or docker exec ...

    If using commands, for python subprocess, you may want to use:
        commands=[
            '-c', 'Put All Bash Command Here no escaped single quotes'
        ]

    If using this to write to a .sh script, you will want commands to be
        [
            '-c', '\'cd somewhere && ./run_thing.sh\''
        ]
    '''

    ret = [
        'docker', docker_command, '--rm', f'--cpus={max(1, (min(threads, get_thread_budget())))}'
    ]
    if user and not IS_WINDOWS:
        ret.extend([
            '--user', user
        ])
    for k,v in env_dict.items():
        ret.extend(['-e', f'"{strip_outer_quotes(k)}={strip_outer_quotes(v)}"'])
    for v in ro_volumes:
        assert isinstance(v, (str, Path, tuple)), f'{type(v)=} {v=} must be Path|str|tuple'
        if isinstance(v, tuple):
            assert len(v) == 2, f'Need src, dst in ro_volumes entry {v}'
            ret.extend(['-v', f'{v[0]}:{v[1]}:ro'])
        else:
            ret.extend(['-v', f'{v}:{v}:ro'])
    for v in rw_volumes:
        assert isinstance(v, (str, Path, tuple)), f'{type(v)=} {v=} must be Path|str|tuple'
        if isinstance(v, tuple):
            assert len(v) == 2, f'Need src, dst in rw_volumes entry {v}'
            ret.extend(['-v', f'{v[0]}:{v[1]}:rw'])
        else:
            ret.extend(['-v', f'{v}:{v}:rw'])

    if entrypoint:
        ret.extend([
            '--entrypoint', entrypoint
        ])

    assert image, f'docker {image=} is missing, must set'
    ret.append(image)

    if commands:
        ret.extend(commands)

    return ret


def check_image_available(image: str) -> bool:
    '''subprocess check if the docker image has been pulled and present on this system'''

    result = subprocess.run(["docker", "image", "inspect", image],
                            capture_output=True, check=False)
    return result.returncode == 0


def run_docker_for_version_info( # pylint: disable=dangerous-default-value
        image: str,
        docker_command: str = 'run', # or 'exec'
        user: str = '',
        env_dict: dict = {},
        entrypoint: str = '/bin/bash',
        commands: list = []
) -> list:
    '''Returns list (suitable for subprocess) of: docker run ... or docker exec ...

    Avoids volume mapping, runs 'docker run' with --rm, runs as default (root) user'''

    cmd_list = get_docker_run_command(
        image=image, docker_command=docker_command,
        user=user, env_dict=env_dict, entrypoint=entrypoint,
        commands=commands
    )
    ret = subprocess.run(cmd_list, capture_output=True, check=False)
    if ret:
        return ret.stdout.decode('utf-8', errors='replace').strip()
    return ''



'''Because > 1 tools use the exe 'code', I don't want to run

`code --list-extensions --show-versionsvsim -version` for every tool that needs
to check if they exist in VScode land.

Instead, eda.py can call this once, and then query if the VScode extension exists when
running opencos.eda.auto_tool_setup(..)
'''

import subprocess

from opencos.files import safe_shutil_which
from opencos.util import debug
from opencos.utils.macos_utils import IS_MACOS, get_macos_tool_info

VSCODE_PATH_INIT_RUN = False
INIT_HAS_RUN = False # pylint: disable=invalid-name
EXTENSIONS = {} # dict of {name: version} for VScode extensions of name

vscode_path = '' # update for macos during init() # pylint: disable=invalid-name

def get_vscode_path() -> str:
    '''Sets global vscode_path if it hasn't been set, returns it'''
    global VSCODE_PATH_INIT_RUN # pylint: disable=global-statement
    global vscode_path          # pylint: disable=global-statement

    if VSCODE_PATH_INIT_RUN:
        return vscode_path

    VSCODE_PATH_INIT_RUN = True

    if vscode_path:
        return vscode_path

    vscode_path = safe_shutil_which('code')

    if not vscode_path and IS_MACOS:
        data = get_macos_tool_info('Visual Studio Code', prefer_exes=['code'])
        if exe := data.get('bin', ''):
            vscode_path = exe

    return vscode_path


vscode_path = get_vscode_path()


def init() -> None:
    '''Sets INIT_HAS_RUN=True (only runs once) and one of TOOL_IS[tool] = True'''
    global INIT_HAS_RUN # pylint: disable=global-statement

    if INIT_HAS_RUN:
        return

    INIT_HAS_RUN = True

    _ = get_vscode_path()

    if not vscode_path:
        return

    proc = None
    try:
        proc = subprocess.run([vscode_path, '--list-extensions', '--show-versions'],
                              capture_output=True, check=False)
    except Exception as e:
        debug(f'vscode --list-extensions --show-versions: exception {e}')

    if proc is None or proc.returncode != 0:
        return

    for line in proc.stdout.decode('utf-8', errors='replace').split('\n'):
        if '@' in line:
            parts = line.split('@')
            if parts[0] and parts[1]:
                EXTENSIONS[parts[0]] = parts[1]
